"""Decoder layer.

The decoder:
- perform a single message-passing step on mesh2grid using a classical interaction network.
- add a residual connection to the grid nodes.
"""

import torch

from graph_weather.models.gencast.layers.modules import MLP, InteractionNetwork


class Decoder(torch.nn.Module):
    """GenCast's decoder."""

    def __init__(
        self,
        edges_dim: int,
        output_dim: int,
        hidden_dims: list[int],
        activation_layer: torch.nn.Module = torch.nn.ReLU,
        use_layer_norm: bool = True,
    ):
        """Initialize the Decoder.

        Args:
            edges_dim (int): dimension of edges' features.
            output_dim (int): dimension of final output.
            hidden_dims (list[int]): hidden dimensions of internal MLPs.
            activation_layer (torch.nn.Module, optional): activation function of internal MLPs.
                Defaults to torch.nn.ReLU.
            use_layer_norm (bool, optional): if true add a LayerNorm at the end of each MLP.
                Defaults to True.
        """
        super().__init__()

        # All the MLPs in GenCast have same hidden and output dims. Hence, the embedding latent
        # dimension and the MLPs' output dimension are the same. Moreover, for simplicity, we will
        # ask the hidden dims just once for each MLP in a module: we don't need to specify them
        # individually as arguments, even if the MLPs could have different roles.
        self.latent_dim = hidden_dims[-1]

        # Embedders
        self.edges_mlp = MLP(
            input_dim=edges_dim,
            hidden_dims=hidden_dims,
            activation_layer=activation_layer,
            use_layer_norm=use_layer_norm,
            bias=True,
            activate_final=False,
        )

        # Message Passing
        self.gnn = InteractionNetwork(
            sender_dim=self.latent_dim,
            receiver_dim=self.latent_dim,
            edge_attr_dim=self.latent_dim,
            hidden_dims=hidden_dims,
            use_layer_norm=use_layer_norm,
            activation_layer=activation_layer,
        )

        # Final grid nodes update
        self.grid_mlp_final = MLP(
            input_dim=self.latent_dim,
            hidden_dims=hidden_dims[:-1] + [output_dim],
            activation_layer=activation_layer,
            use_layer_norm=use_layer_norm,
            bias=True,
            activate_final=False,
        )

    def forward(
        self,
        input_mesh_nodes: torch.Tensor,
        input_grid_nodes: torch.Tensor,
        input_edge_attr: torch.Tensor,
        edge_index: torch.Tensor,
    ) -> torch.Tensor:
        """Forward pass.

        Args:
            input_mesh_nodes (torch.Tensor): mesh nodes' features.
            input_grid_nodes (torch.Tensor): grid nodes' features.
            input_edge_attr (torch.Tensor): grid2mesh edges' features.
            edge_index (torch.Tensor): edge index tensor.

        Returns:
            torch.Tensor: output grid nodes.
        """
        if not (
            input_grid_nodes.shape[-1] == self.latent_dim
            and input_mesh_nodes.shape[-1] == self.latent_dim
        ):
            raise ValueError(
                "The dimension of grid nodes and mesh nodes' features must be "
                "equal to the last hidden dimension."
            )

        # Embedding
        edges_emb = self.edges_mlp(input_edge_attr)

        # Message-passing + residual connection
        latent_grid_nodes = input_grid_nodes + self.gnn(
            x=(input_mesh_nodes, input_grid_nodes),
            edge_index=edge_index,
            edge_attr=edges_emb,
        )

        # Update grid nodes
        latent_grid_nodes = self.grid_mlp_final(latent_grid_nodes)

        return latent_grid_nodes



"""Diffusion sampler"""

import math

import torch

from graph_weather.models.gencast import Denoiser
from graph_weather.models.gencast.utils.noise import generate_isotropic_noise


class Sampler:
    """Sampler for the denoiser.

    The sampler consists in the second-order DPMSolver++2S solver (Lu et al., 2022), augmented with
    the stochastic churn (again making use of the isotropic noise) and noise inflation techniques
    used in Karras et al. (2022) to inject further stochasticity into the sampling process. In
    conditioning on previous timesteps it follows the Conditional Denoising Estimator approach
    outlined and motivated by Batzolis et al. (2021).
    """

    def __init__(
        self,
        S_noise: float = 1.05,
        S_tmin: float = 0.75,
        S_tmax: float = 80.0,
        S_churn: float = 2.5,
        r: float = 0.5,
        sigma_max: float = 80.0,
        sigma_min: float = 0.03,
        rho: float = 7,
        num_steps: int = 20,
    ):
        """Initialize the sampler.

        Args:
            S_noise (float): noise inflation parameter. Defaults to 1.05.
            S_tmin (float): minimum noise for sampling. Defaults to 0.75.
            S_tmax (float): maximum noise for sampling. Defaults to 80.
            S_churn (float): stochastic churn rate. Defaults to 2.5.
            r (float): _description_. Defaults to 0.5.
            sigma_max (float): maximum value of sigma for sigma's distribution. Defaults to 80.
            sigma_min (float): minimum value of sigma for sigma's distribution. Defaults to 0.03.
            rho (float): exponent of the sigma's distribution. Defaults to 7.
            num_steps (int): number of timesteps during sampling. Defaults to 20.
        """
        self.S_noise = S_noise
        self.S_tmin = S_tmin
        self.S_tmax = S_tmax
        self.S_churn = S_churn
        self.r = r
        self.num_steps = num_steps

        self.sigma_max = sigma_max
        self.sigma_min = sigma_min
        self.rho = rho

    def _sigmas_fn(self, u):
        return (
            self.sigma_max ** (1 / self.rho)
            + u * (self.sigma_min ** (1 / self.rho) - self.sigma_max ** (1 / self.rho))
        ) ** self.rho

    @torch.no_grad()
    def sample(self, denoiser: Denoiser, prev_inputs: torch.Tensor):
        """Generate a sample from random noise for the given inputs.

        Args:
            denoiser (Denoiser): the denoiser model.
            prev_inputs (torch.Tensor): previous two timesteps.

        Returns:
            torch.Tensor: normalized residuals predicted.
        """
        device = prev_inputs.device

        time_steps = torch.arange(0, self.num_steps).to(device) / (self.num_steps - 1)
        sigmas = self._sigmas_fn(time_steps)

        batch_ones = torch.ones(1, 1).to(device)

        # initialize noise
        x = sigmas[0] * torch.tensor(
            generate_isotropic_noise(
                num_lon=denoiser.num_lon,
                num_lat=denoiser.num_lat,
                num_samples=denoiser.output_features_dim,
            )
        ).unsqueeze(0).to(device)

        for i in range(len(sigmas) - 1):
            # stochastic churn from Karras et al. (Alg. 2)
            gamma = (
                min(self.S_churn / self.num_steps, math.sqrt(2) - 1)
                if self.S_tmin <= sigmas[i] <= self.S_tmax
                else 0.0
            )
            # noise inflation from Karras et al. (Alg. 2)
            noise = self.S_noise * torch.tensor(
                generate_isotropic_noise(
                    num_lon=denoiser.num_lon,
                    num_lat=denoiser.num_lat,
                    num_samples=denoiser.output_features_dim,
                )
            )
            noise = noise.unsqueeze(0).to(device)

            sigma_hat = sigmas[i] * (gamma + 1)
            if gamma > 0:
                x = x + (sigma_hat**2 - sigmas[i] ** 2) ** 0.5 * noise
            denoised = denoiser(x, prev_inputs, sigma_hat * batch_ones)

            if i == len(sigmas) - 2:
                # final Euler step
                d = (x - denoised) / sigma_hat
                x = x + d * (sigmas[i + 1] - sigma_hat)
            else:
                # DPMSolver++2S  step (Alg. 1 in Lu et al.) with alpha_t=1.
                # t_{i-1} is t_hat because of stochastic churn!
                lambda_hat = -torch.log(sigma_hat)
                lambda_next = -torch.log(sigmas[i + 1])
                h = lambda_next - lambda_hat
                lambda_mid = lambda_hat + self.r * h
                sigma_mid = torch.exp(-lambda_mid)

                u = sigma_mid / sigma_hat * x - (torch.exp(-self.r * h) - 1) * denoised
                denoised_2 = denoiser(u, prev_inputs, sigma_mid * batch_ones)
                D = (1 - 1 / (2 * self.r)) * denoised + 1 / (2 * self.r) * denoised_2
                x = sigmas[i + 1] / sigma_hat * x - (torch.exp(-h) - 1) * D

        return x



"""
Functions for building GNN

This code is taken from https://github.com/CCSI-Toolset/MGN which is available under the
US Government License
"""

from typing import Optional, Tuple

import torch
from torch import cat, nn
from torch.utils.checkpoint import checkpoint
from torch_geometric.nn import MetaLayer
from torch_scatter import scatter_sum


class MLP(nn.Module):
    """MLP for graph processing"""

    def __init__(
        self,
        in_dim: int,
        out_dim: int = 128,
        hidden_dim: int = 128,
        hidden_layers: int = 2,
        norm_type: Optional[str] = "LayerNorm",
        use_checkpointing: bool = False,
    ):
        """
        MLP

        Args:
            in_dim: Input dimension
            out_dim: Output dimension
            hidden_dim: Number of nodes in hidden layer
            hidden_layers: Number of hidden layers
            norm_type: Normalization type one of 'LayerNorm', 'GraphNorm',
                'InstanceNorm', 'BatchNorm', 'MessageNorm', or None
            use_checkpointing: Whether to use gradient checkpointing or not
        """

        super(MLP, self).__init__()
        self.use_checkpointing = use_checkpointing

        layers = [nn.Linear(in_dim, hidden_dim), nn.ReLU()]
        for _ in range(hidden_layers - 1):
            layers += [nn.Linear(hidden_dim, hidden_dim), nn.ReLU()]
        layers.append(nn.Linear(hidden_dim, out_dim))

        if norm_type is not None:
            assert norm_type in [
                "LayerNorm",
                "GraphNorm",
                "InstanceNorm",
                "BatchNorm",
                "MessageNorm",
            ]
            norm_layer = getattr(nn, norm_type)
            layers.append(norm_layer(out_dim))

        self.model = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Compute the MLP

        Args:
            x: Node or edge features

        Returns:
            The transformed tensor
        """
        if self.use_checkpointing:
            out = checkpoint(self.model, x, use_reentrant=False)
        else:
            out = self.model(x)
        return out


#############################

# issue with MessagePassing class:
# Only node features are updated after MP iterations
# Need to use MetaLayer to also allow edge features to update


class EdgeProcessor(nn.Module):
    """EdgeProcessor"""

    def __init__(
        self,
        in_dim_node: int = 128,
        in_dim_edge: int = 128,
        hidden_dim: int = 128,
        hidden_layers: int = 2,
        norm_type: str = "LayerNorm",
    ):
        """
        Edge processor

        Args:
            in_dim_node: Input node feature dimension
            in_dim_edge: Input edge feature dimension
            hidden_dim: Number of nodes in hidden layers
            hidden_layers: Number of hidden layers
            norm_type: Normalization type
                one of 'LayerNorm', 'GraphNorm', 'InstanceNorm', 'BatchNorm', 'MessageNorm', or None
        """

        super(EdgeProcessor, self).__init__()
        self.edge_mlp = MLP(
            2 * in_dim_node + in_dim_edge, in_dim_edge, hidden_dim, hidden_layers, norm_type
        )

    def forward(
        self, src: torch.Tensor, dest: torch.Tensor, edge_attr: torch.Tensor, u=None, batch=None
    ) -> torch.Tensor:
        """
        Compute the edge part of the message passing

        Args:
            src: Source node tensor
            dest: Destination node tensor
            edge_attr: Edge attributes
            u: Global attributes, ignored
            batch: Batch Ids, ignored

        Returns:
            The updated edge attributes
        """
        out = cat(
            [src, dest, edge_attr], -1
        )  # concatenate source node, destination node, and edge embeddings
        out = self.edge_mlp(out)
        out += edge_attr  # residual connection

        return out


class NodeProcessor(nn.Module):
    """NodeProcessor"""

    def __init__(
        self,
        in_dim_node: int = 128,
        in_dim_edge: int = 128,
        hidden_dim: int = 128,
        hidden_layers: int = 2,
        norm_type: str = "LayerNorm",
    ):
        """
        Node Processor

        Args:
            in_dim_node: Input node feature dimension
            in_dim_edge: Input edge feature dimension
            hidden_dim: Number of nodes in hidden layer
            hidden_layers: Number of hidden layers
            norm_type: Normalization type
                one of 'LayerNorm', 'GraphNorm', 'InstanceNorm', 'BatchNorm', 'MessageNorm', or None
        """

        super(NodeProcessor, self).__init__()
        self.node_mlp = MLP(
            in_dim_node + in_dim_edge, in_dim_node, hidden_dim, hidden_layers, norm_type
        )

    def forward(
        self, x: torch.Tensor, edge_index: torch.Tensor, edge_attr: torch.Tensor, u=None, batch=None
    ) -> torch.Tensor:
        """
        Compute the node feature updates in message passing

        Args:
            x: Input nodes
            edge_index: Edge indicies in COO format
            edge_attr: Edge attributes
            u: Global attributes, ignored
            batch: Batch IDX, ignored

        Returns:
            torch.Tensor with updated node attributes
        """
        row, col = edge_index
        scatter_dim = 0
        output_size = x.size(scatter_dim)
        # aggregate edge message by target
        out = scatter_sum(edge_attr, col, dim=scatter_dim, dim_size=output_size)
        out = cat([x, out], dim=-1)
        out = self.node_mlp(out)
        out += x  # residual connection

        return out


def build_graph_processor_block(
    in_dim_node: int = 128,
    in_dim_edge: int = 128,
    hidden_dim_node: int = 128,
    hidden_dim_edge: int = 128,
    hidden_layers_node: int = 2,
    hidden_layers_edge: int = 2,
    norm_type: str = "LayerNorm",
) -> torch.nn.Module:
    """
    Build the Graph Net Block

    Args:
        in_dim_node: Input node feature dimension
        in_dim_edge: Input edge feature dimension
        hidden_dim_node: Number of nodes in hidden layer for graph node processing
        hidden_dim_edge: Number of nodes in hidden layer for graph edge processing
        hidden_layers_node: Number of hidden layers for node processing
        hidden_layers_edge: Number of hidden layers for edge processing
        norm_type: Normalization type
                one of 'LayerNorm', 'GraphNorm', 'InstanceNorm', 'BatchNorm', 'MessageNorm', or None
    Returns:
        torch.nn.Module for the graph processing block
    """

    return MetaLayer(
        edge_model=EdgeProcessor(
            in_dim_node, in_dim_edge, hidden_dim_edge, hidden_layers_edge, norm_type
        ),
        node_model=NodeProcessor(
            in_dim_node, in_dim_edge, hidden_dim_node, hidden_layers_node, norm_type
        ),
    )


class GraphProcessor(nn.Module):
    """Overall graph processor"""

    def __init__(
        self,
        mp_iterations: int = 15,
        in_dim_node: int = 128,
        in_dim_edge: int = 128,
        hidden_dim_node: int = 128,
        hidden_dim_edge: int = 128,
        hidden_layers_node: int = 2,
        hidden_layers_edge: int = 2,
        norm_type: str = "LayerNorm",
    ):
        """
        Graph Processor

        Args:
            mp_iterations: number of message-passing iterations (graph processor blocks)
            in_dim_node: Input node feature dimension
            in_dim_edge: Input edge feature dimension
            hidden_dim_node: Number of nodes in hidden layers for node processing
            hidden_dim_edge: Number of nodes in hidden layers for edge processing
            hidden_layers_node: Number of hidden layers for node processing
            hidden_layers_edge: Number of hidden layers for edge processing
            norm_type: Normalization type
                one of 'LayerNorm', 'GraphNorm', 'InstanceNorm', 'BatchNorm', 'MessageNorm', or None
        """

        super(GraphProcessor, self).__init__()

        self.blocks = nn.ModuleList()
        for _ in range(mp_iterations):
            self.blocks.append(
                build_graph_processor_block(
                    in_dim_node,
                    in_dim_edge,
                    hidden_dim_node,
                    hidden_dim_edge,
                    hidden_layers_node,
                    hidden_layers_edge,
                    norm_type,
                )
            )

    def forward(
        self, x: torch.Tensor, edge_index: torch.Tensor, edge_attr: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Compute updates to the graph in message passing method

        Args:
            x: Input nodes
            edge_index: Edge indicies in COO format
            edge_attr: Edge attributes

        Returns:
            Updated nodes and edge attributes
        """
        for block in self.blocks:
            x, edge_attr, _ = block(x, edge_index, edge_attr)

        return x, edge_attr



"""Processor layer.

The processor:
- compute a sequence of transformer blocks applied to the mesh.
- condition on the noise level.
"""

import torch

from graph_weather.models.gencast.layers.modules import MLP, CondTransformerBlock, FourierEmbedding

try:
    from graph_weather.models.gencast.layers.experimental import SparseTransformer

    has_dgl = True
except ImportError:
    has_dgl = False


class Processor(torch.nn.Module):
    """GenCast's Processor

    The Processor is a sequence of transformer blocks conditioned on noise level. If the graph has
    many edges, setting sparse=True may perform better in terms of memory and speed. Note that
    sparse=False uses PyG as the backend, while sparse=True uses DGL. The two implementations are
    not exactly equivalent: the former is described in the paper "Masked Label Prediction: Unified
    Message Passing Model for Semi-Supervised Classification" and can also handle edge features,
    while the latter is a classical transformer that performs multi-head attention utilizing the
    mask's sparsity and does not include edge features in the computations.

    Note: The GenCast paper does not provide specific details regarding the implementation of the
    transformer architecture for graphs.
    """

    def __init__(
        self,
        latent_dim: int,
        hidden_dims: list[int],
        num_blocks: int,
        num_heads: int,
        num_frequencies: int,
        base_period: int,
        noise_emb_dim: int,
        edges_dim: int | None = None,
        activation_layer: torch.nn.Module = torch.nn.ReLU,
        use_layer_norm: bool = True,
        sparse: bool = False,
    ):
        """Initialize the Processor.

        Args:
            latent_dim (int): dimension of nodes' features.
            hidden_dims (list[int]): hidden dimensions of internal MLPs.
            num_blocks (int): number of transformer blocks.
            num_heads (int): number of heads for multi-head attention.
            num_frequencies (int): number of frequencies for the noise Fourier embedding.
            base_period (int): base period for the noise Fourier embedding.
            noise_emb_dim (int): dimension of output of noise embedding.
            edges_dim (int, optional): dimension of edges' features. If None does not uses edges
                features in TransformerConv. Defaults to None.
            activation_layer (torch.nn.Module): activation function of internal MLPs.
                Defaults to torch.nn.ReLU.
            use_layer_norm (bool): if true add a LayerNorm at the end of the embedding MLP.
                Defaults to True.
            sparse (bool): if true use DGL as backend (experimental). Defaults to False.
        """
        super().__init__()
        self.latent_dim = latent_dim
        if latent_dim % num_heads != 0:
            raise ValueError("The latent dimension should be divisible by the number of heads.")

        # Embedders
        self.fourier_embedder = FourierEmbedding(
            output_dim=noise_emb_dim, num_frequencies=num_frequencies, base_period=base_period
        )

        self.edges_dim = edges_dim
        if edges_dim is not None:
            self.edges_mlp = MLP(
                input_dim=edges_dim,
                hidden_dims=hidden_dims,
                activation_layer=activation_layer,
                use_layer_norm=use_layer_norm,
                bias=True,
                activate_final=False,
            )

        # Tranformers Blocks
        self.cond_transformers = torch.nn.ModuleList()
        if not sparse:
            for _ in range(num_blocks - 1):
                # concatenating multi-head attention
                self.cond_transformers.append(
                    CondTransformerBlock(
                        conditioning_dim=noise_emb_dim,
                        input_dim=latent_dim,
                        output_dim=latent_dim // num_heads,
                        edges_dim=hidden_dims[-1] if (edges_dim is not None) else None,
                        num_heads=num_heads,
                        concat=True,
                        beta=True,
                        activation_layer=activation_layer,
                    )
                )

            # averaging multi-head attention
            self.cond_transformers.append(
                CondTransformerBlock(
                    conditioning_dim=noise_emb_dim,
                    input_dim=latent_dim,
                    output_dim=latent_dim,
                    edges_dim=hidden_dims[-1] if (edges_dim is not None) else None,
                    num_heads=num_heads,
                    concat=False,
                    beta=True,
                    activation_layer=None,
                )
            )
        else:
            if not has_dgl:
                raise ValueError("Please install DGL to use sparsity.")

            for _ in range(num_blocks):
                # concatenating multi-head attention
                self.cond_transformers.append(
                    SparseTransformer(
                        conditioning_dim=noise_emb_dim,
                        input_dim=latent_dim,
                        output_dim=latent_dim,
                        num_heads=num_heads,
                        activation_layer=activation_layer,
                    )
                )
                # do we really need averaging for last block?

    def _check_args(self, latent_mesh_nodes, noise_levels, input_edge_attr):
        if not latent_mesh_nodes.shape[-1] == self.latent_dim:
            raise ValueError(
                "The dimension of the mesh nodes is different from the latent dimension provided at"
                " initialization."
            )

        if not latent_mesh_nodes.shape[0] == noise_levels.shape[0]:
            raise ValueError(
                "The number of noise levels and mesh nodes should be the same, but got "
                f"{latent_mesh_nodes.shape[0]} and {noise_levels.shape[0]}. Eventually repeat the "
                " noise level for each node in the same batch."
            )

        if (input_edge_attr is not None) and (self.edges_dim is None):
            raise ValueError("To use input_edge_attr initialize the processor with edges_dim.")

    def forward(
        self,
        latent_mesh_nodes: torch.Tensor,
        edge_index: torch.Tensor,
        noise_levels: torch.Tensor,
        input_edge_attr: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """Forward pass.

        Args:
            latent_mesh_nodes (torch.Tensor): mesh nodes' features.
            edge_index (torch.Tensor): edge index tensor.
            noise_levels (torch.Tensor): log-noise levels.
            input_edge_attr (torch.Tensor, optional): mesh edges' features.

        Returns:
            torch.Tensor: latent mesh nodes.
        """
        self._check_args(latent_mesh_nodes, noise_levels, input_edge_attr)

        # embedding
        noise_emb = self.fourier_embedder(noise_levels)

        if self.edges_dim is not None:
            edges_emb = self.edges_mlp(input_edge_attr)
        else:
            edges_emb = None

        # apply transformer blocks
        for cond_transformer in self.cond_transformers:
            latent_mesh_nodes = cond_transformer(
                x=latent_mesh_nodes,
                edge_index=edge_index,
                cond_param=noise_emb,
                edge_attr=edges_emb,
            )

        return latent_mesh_nodes



"""Utils for batching graphs."""

import torch


def batch(senders, edge_index, edge_attr=None, batch_size=1):
    """Build big batched graph.

    Returns nodes and edges of a big graph with batch_size disconnected copies of the original
    graph, with features shape [(b n) f].

    Args:
        senders (torch.Tensor): nodes' features.
        edge_index (torch.Tensor): edge index tensor.
        edge_attr (torch.Tensor, optional): edge attributes tensor, if None returns None.
            Defaults to None.
        batch_size (int): batch size. Defaults to 1.

    Returns:
        batched_senders, batched_edge_index, batched_edge_attr
    """
    ns = senders.shape[0]
    batched_senders = senders
    batched_edge_attr = edge_attr
    batched_edge_index = edge_index

    for i in range(1, batch_size):
        batched_senders = torch.cat([batched_senders, senders], dim=0)
        batched_edge_index = torch.cat([batched_edge_index, edge_index + i * ns], dim=1)

        if edge_attr is not None:
            batched_edge_attr = torch.cat([batched_edge_attr, edge_attr], dim=0)

    return batched_senders, batched_edge_index, batched_edge_attr


def hetero_batch(senders, receivers, edge_index, edge_attr=None, batch_size=1):
    """Build big batched heterogenous graph.

    Returns nodes and edges of a big graph with batch_size disconnected copies of the original
    graph, with features shape [(b n) f].

    Args:
        senders (torch.Tensor): senders' features.
        receivers (torch.Tensor): receivers' features.
        edge_index (torch.Tensor): edge index tensor.
        edge_attr (torch.Tensor, optional): edge attributes tensor, if None returns None.
            Defaults to None.
        batch_size (int): batch size. Defaults to 1.

    Returns:
        batched_senders, batched_edge_index, batched_edge_attr
    """
    ns = senders.shape[0]
    nr = receivers.shape[0]
    nodes_shape = torch.tensor([[ns], [nr]]).to(edge_index)
    batched_senders = senders
    batched_receivers = receivers
    batched_edge_attr = edge_attr
    batched_edge_index = edge_index

    for i in range(1, batch_size):
        batched_senders = torch.cat([batched_senders, senders], dim=0)
        batched_receivers = torch.cat([batched_receivers, receivers], dim=0)
        batched_edge_index = torch.cat([batched_edge_index, edge_index + i * nodes_shape], dim=1)
        if edge_attr is not None:
            batched_edge_attr = torch.cat([batched_edge_attr, edge_attr], dim=0)

    return batched_senders, batched_receivers, batched_edge_index, batched_edge_attr



"""Experimental sparse transformer using DGL sparse."""

import dgl.sparse as dglsp
import torch
import torch.nn as nn

from graph_weather.models.gencast.layers.modules import ConditionalLayerNorm


class SparseAttention(nn.Module):
    """Sparse Multi-head Attention Module"""

    def __init__(self, input_dim=512, output_dim=512, num_heads=4):
        """Initialize Sparse MultiHead attention module.

        Args:
            input_dim (int): input dimension. Defaults to 512.
            output_dim (int): output dimension. Defaults to 512.
            num_heads (int): number of heads. Output dimension should be divisible by num_heads.
                Defaults to 4.
        """
        super().__init__()
        if output_dim % num_heads:
            raise ValueError("Output dimension should be divisible by the number of heads.")

        self.hidden_size = output_dim
        self.num_heads = num_heads
        self.head_dim = output_dim // num_heads
        self.scaling = self.head_dim**-0.5

        self.q_proj = nn.Linear(input_dim, output_dim)
        self.k_proj = nn.Linear(input_dim, output_dim)
        self.v_proj = nn.Linear(input_dim, output_dim)
        self.out_proj = nn.Linear(output_dim, output_dim)

    def forward(self, x: torch.Tensor, adj: dglsp.SparseMatrix):
        """Forward pass of SparseMHA.

        Args:
            x (torch.Tensor): input tensor.
            adj (SparseMatrix): adjacency matrix in DGL SparseMatrix format.

        Returns:
            y (tensor): output of MultiHead attention.
        """
        N = len(x)
        # computing query,key and values.
        q = self.q_proj(x).reshape(N, self.head_dim, self.num_heads)  # (dense) [N, dh, nh]
        k = self.k_proj(x).reshape(N, self.head_dim, self.num_heads)  # (dense) [N, dh, nh]
        v = self.v_proj(x).reshape(N, self.head_dim, self.num_heads)  # (dense) [N, dh, nh]
        # scaling query
        q *= self.scaling

        # sparse-dense-dense product
        attn = dglsp.bsddmm(adj, q, k.transpose(1, 0))  # (sparse) [N, N, nh]

        # sparse softmax (by default applies on the last sparse dimension).
        attn = attn.softmax()  # (sparse) [N, N, nh]

        # sparse-dense multiplication
        out = dglsp.bspmm(attn, v)  # (dense) [N, dh, nh]
        return self.out_proj(out.reshape(N, -1))


class SparseTransformer(nn.Module):
    """A single transformer block for graph neural networks.

    This module implements a single transformer block with a sparse attention mechanism.
    """

    def __init__(
        self,
        conditioning_dim: int,
        input_dim: int,
        output_dim: int,
        num_heads: int,
        activation_layer: torch.nn.Module = nn.ReLU,
        norm_first: bool = True,
    ):
        """Initialize SparseTransformer module.

        Args:
            conditioning_dim (int, optional): dimension of the conditioning parameter. If None the
                layer normalization will not be applied.
            input_dim (int): dimension of the input features.
            output_dim (int): dimension of the output features.
            edges_dim (int): dimension of the edge features.
            num_heads (int): number of heads for multi-head attention.
            activation_layer (torch.nn.Module): activation function applied before
                returning the output.
            norm_first (bool): if True apply layer normalization before attention. Defaults to True.
        """
        super().__init__()

        # initialize multihead sparse attention.
        self.sparse_attention = SparseAttention(
            input_dim=input_dim, output_dim=output_dim, num_heads=num_heads
        )

        # initialize mlp
        self.activation = activation_layer()
        self.mlp = nn.Sequential(
            nn.Linear(output_dim, output_dim), self.activation, nn.Linear(output_dim, output_dim)
        )

        # initialize conditional layer normalization
        self.cond_norm_1 = ConditionalLayerNorm(
            conditioning_dim=conditioning_dim, features_dim=output_dim
        )
        self.cond_norm_2 = ConditionalLayerNorm(
            conditioning_dim=conditioning_dim, features_dim=output_dim
        )

        self.norm_first = norm_first

    def forward(
        self,
        x: torch.Tensor,
        edge_index: torch.Tensor,
        cond_param: torch.Tensor,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        """Apply SparseTransformer to input.

        Input and conditioning parameter must have same batch size.

        Args:
            x (torch.Tensor): tensor containing nodes features.
            edge_index (torch.Tensor): edge index tensor.
            cond_param (torch.Tensor): conditioning parameter.
            *args: ignored by the module.
            **kwargs: ignored by the module.

        """
        if self.norm_first:
            x1 = self.cond_norm_1(x, cond_param)
            x = x + self.sparse_attention(
                x=x1, adj=dglsp.spmatrix(indices=edge_index, shape=(x.shape[0], x.shape[0]))
            )
        else:
            x = x + self.sparse_attention(
                x=x, adj=dglsp.spmatrix(indices=edge_index, shape=(x.shape[0], x.shape[0]))
            )
            x = self.cond_norm_1(x, cond_param)

        if self.norm_first:
            x2 = self.cond_norm_2(x, cond_param)
            x = x + self.mlp(x2)
        else:
            x = x + self.mlp(x)
            x = self.cond_norm_2(x, cond_param)
        return x



import numpy as np
import torch
import pytest
from packaging.version import Version

from torch_geometric.transforms import TwoHop

from graph_weather.models.gencast.utils.noise import (
    generate_isotropic_noise,
    sample_noise_level,
)
from graph_weather.models.gencast import GraphBuilder, WeightedMSELoss, Denoiser, Sampler
from graph_weather.models.gencast.layers.modules import FourierEmbedding


def test_gencast_noise():
    num_lon = 360
    num_lat = 180
    num_samples = 5
    target_residuals = np.zeros((num_lon, num_lat, num_samples))
    noise_level = sample_noise_level()
    noise = generate_isotropic_noise(
        num_lon=num_lon, num_lat=num_lat, num_samples=target_residuals.shape[-1]
    )
    corrupted_residuals = target_residuals + noise_level * noise
    assert corrupted_residuals.shape == target_residuals.shape
    assert not np.isnan(corrupted_residuals).any()

    num_lon = 360
    num_lat = 181
    num_samples = 5
    target_residuals = np.zeros((num_lon, num_lat, num_samples))
    noise_level = sample_noise_level()
    noise = generate_isotropic_noise(
        num_lon=num_lon, num_lat=num_lat, num_samples=target_residuals.shape[-1]
    )
    corrupted_residuals = target_residuals + noise_level * noise
    assert corrupted_residuals.shape == target_residuals.shape
    assert not np.isnan(corrupted_residuals).any()

    num_lon = 100
    num_lat = 100
    num_samples = 5
    target_residuals = np.zeros((num_lon, num_lat, num_samples))
    noise_level = sample_noise_level()
    noise = generate_isotropic_noise(
        num_lon=num_lon, num_lat=num_lat, num_samples=target_residuals.shape[-1], isotropic=False
    )
    corrupted_residuals = target_residuals + noise_level * noise
    assert corrupted_residuals.shape == target_residuals.shape
    assert not np.isnan(corrupted_residuals).any()


def test_gencast_graph():
    grid_lat = np.arange(-90, 90, 1)
    grid_lon = np.arange(0, 360, 1)
    graphs = GraphBuilder(grid_lon=grid_lon, grid_lat=grid_lat, splits=4, num_hops=8)

    # compare khop sparse implementation with pyg.
    transform = TwoHop()
    khop_mesh_graph_pyg = graphs.mesh_graph
    for i in range(3):  # 8-hop mesh
        khop_mesh_graph_pyg = transform(khop_mesh_graph_pyg)

    assert graphs.mesh_graph.x.shape[0] == 2562
    assert graphs.g2m_graph["grid_nodes"].x.shape[0] == 360 * 180
    assert graphs.m2g_graph["mesh_nodes"].x.shape[0] == 2562
    assert not torch.isnan(graphs.mesh_graph.edge_attr).any()
    assert graphs.khop_mesh_graph.x.shape[0] == 2562
    assert torch.allclose(graphs.khop_mesh_graph.x, khop_mesh_graph_pyg.x)
    assert torch.allclose(graphs.khop_mesh_graph.edge_index, khop_mesh_graph_pyg.edge_index)


def test_gencast_loss():
    grid_lat = torch.arange(-90, 90, 1)
    grid_lon = torch.arange(0, 360, 1)
    pressure_levels = torch.tensor(
        [50.0, 100.0, 150.0, 200.0, 250, 300, 400, 500, 600, 700, 850, 925, 1000.0]
    )
    single_features_weights = torch.tensor([1, 0.1, 0.1, 0.1, 0.1])
    num_atmospheric_features = 6
    batch_size = 3
    features_dim = len(pressure_levels) * num_atmospheric_features + len(single_features_weights)

    loss = WeightedMSELoss(
        grid_lat=grid_lat,
        pressure_levels=pressure_levels,
        num_atmospheric_features=num_atmospheric_features,
        single_features_weights=single_features_weights,
    )

    preds = torch.rand((batch_size, len(grid_lon), len(grid_lat), features_dim))
    noise_levels = torch.rand((batch_size, 1))
    targets = torch.rand((batch_size, len(grid_lon), len(grid_lat), features_dim))
    assert loss.forward(preds, noise_levels, targets) is not None


def test_gencast_denoiser():
    grid_lat = np.arange(-90, 90, 1)
    grid_lon = np.arange(0, 360, 1)
    input_features_dim = 10
    output_features_dim = 5
    batch_size = 3

    denoiser = Denoiser(
        grid_lon=grid_lon,
        grid_lat=grid_lat,
        input_features_dim=input_features_dim,
        output_features_dim=output_features_dim,
        hidden_dims=[16, 32],
        num_blocks=3,
        num_heads=4,
        splits=0,
        num_hops=1,
        device=torch.device("cpu"),
    ).eval()

    corrupted_targets = torch.randn((batch_size, len(grid_lon), len(grid_lat), output_features_dim))
    prev_inputs = torch.randn((batch_size, len(grid_lon), len(grid_lat), 2 * input_features_dim))
    noise_levels = torch.rand((batch_size, 1))

    with torch.no_grad():
        preds = denoiser(
            corrupted_targets=corrupted_targets, prev_inputs=prev_inputs, noise_levels=noise_levels
        )

    assert not torch.isnan(preds).any()


def test_gencast_fourier():
    batch_size = 10
    output_dim = 20
    fourier_embedder = FourierEmbedding(output_dim=output_dim, num_frequencies=32, base_period=16)
    t = torch.rand((batch_size, 1))
    assert fourier_embedder(t).shape == (batch_size, output_dim)


def test_gencast_sampler():
    grid_lat = np.arange(-90, 90, 1)
    grid_lon = np.arange(0, 360, 1)
    input_features_dim = 10
    output_features_dim = 5

    denoiser = Denoiser(
        grid_lon=grid_lon,
        grid_lat=grid_lat,
        input_features_dim=input_features_dim,
        output_features_dim=output_features_dim,
        hidden_dims=[16, 32],
        num_blocks=3,
        num_heads=4,
        splits=0,
        num_hops=1,
        device=torch.device("cpu"),
    ).eval()

    prev_inputs = torch.randn((1, len(grid_lon), len(grid_lat), 2 * input_features_dim))

    sampler = Sampler()
    preds = sampler.sample(denoiser, prev_inputs)
    assert not torch.isnan(preds).any()
    assert preds.shape == (1, len(grid_lon), len(grid_lat), output_features_dim)


@pytest.mark.skipif(
    Version(torch.__version__).release != Version("2.3.0").release,
    reason="dgl tests for experimental features only runs with torch 2.3.0",
)
def test_gencast_full():
    # download weights from HF
    denoiser = Denoiser.from_pretrained(
        "openclimatefix/gencast-128x64",
        grid_lon=np.arange(0, 360, 360 / 128),
        grid_lat=np.arange(-90, 90, 180 / 64) + 1 / 2 * 180 / 64,
    )

    # load inputs and targets
    prev_inputs = torch.randn([1, 128, 64, 178])
    target_residuals = torch.randn([1, 128, 64, 83])

    # predict
    sampler = Sampler()
    preds = sampler.sample(denoiser, prev_inputs)

    assert not torch.isnan(preds).any()
    assert preds.shape == target_residuals.shape



"""
The dataloader for IFS analysis.
"""

import numpy as np
import torchvision.transforms as transforms
import xarray as xr
from torch.utils.data import Dataset

IFS_MEAN = {
    "geopotential": 78054.78,
    "specific_humidity": 0.0018220816,
    "temperature": 243.41727,
    "u_component_of_wind": 7.3073797,
    "v_component_of_wind": 0.032221083,
    "vertical_velocity": 0.0058287205,
}

IFS_STD = {
    "geopotential": 59538.875,
    "specific_humidity": 0.0035489395,
    "temperature": 29.211119,
    "u_component_of_wind": 13.777036,
    "v_component_of_wind": 8.867598,
    "vertical_velocity": 0.08577341,
}


class IFSAnalisysDataset(Dataset):
    """
    Dataset for IFSAnalysis.

    Args:
            filepath: path of the dataset.
            features: list of features.
            start_year: initial year. Defaults to 2016.
            end_year: ending year. Defaults to 2022.
    """

    def __init__(self, filepath: str, features: list, start_year: int = 2016, end_year: int = 2022):
        """
        Initialize the dataset object.
        """

        super().__init__()
        assert (
            start_year <= end_year
        ), f"start_year ({start_year}) cannot be greater than end_year ({end_year})."
        assert start_year >= 2016 and start_year <= 2022, "Time data range from 2016 to 2022"
        assert end_year >= 2016 and end_year <= 2022, "Time data range from 2016 to 2022"
        self.data = xr.open_zarr(filepath)
        self.data = self.data.sel(
            time=slice(str(start_year), str(end_year))
        )  # Filter data by start and end years

        self.NWP_features = features

    def __len__(self):
        return len(self.data["time"])

    def __getitem__(self, idx):

        start = self.data.isel(time=idx)
        end = self.data.isel(time=idx + 1)

        # Extract NWP features
        input_data = self._nwp_features_extraction(start)
        output_data = self._nwp_features_extraction(end)

        return (
            (transforms).ToTensor()(input_data).view(-1, input_data.shape[-1]),
            (transforms).ToTensor()(output_data).view(-1, output_data.shape[-1]),
        )

    def _nwp_features_extraction(self, data):
        data_cube = np.stack(
            [
                (data[f"{var}"].values - IFS_MEAN[f"{var}"]) / (IFS_STD[f"{var}"] + 1e-6)
                for var in self.NWP_features
            ],
            axis=-1,
        ).astype(np.float32)

        num_layers, num_lat, num_lon, num_vars = data_cube.shape
        data_cube = data_cube.reshape(num_lat, num_lon, num_vars * num_layers)

        assert not np.isnan(data_cube).any()
        return data_cube



"""Encoders to encode from the input graph to the latent graph

In the original paper the encoder is described as

The Encoder maps from physical data defined on a latitude/longitude grid to abstract latent features
defined on an icosahedron grid.  The Encoder GNN uses a bipartite graph(lat/lon→icosahedron) with
edges only between nodes in the lat/lon grid and nodes in the icosahedron grid. Put another way,
spatial and channel information in the local neighborhood of each icosahedron node is
gathered using connections to nearby lat/lon nodes.

The initial node features are the 78 atmospheric variables described in Section 2.1, plus solar
radiation, orography, land-sea mask, the day-of-year,sin(lat),cos(lat),sin(lon), and cos(lon).
The initial edge features are the positions of the lat/lon nodes connected to each icosahedron node.
These positions are provided in a local coordinate system that is defined relative
to each icosahedron node.

In further notes, they notice that there is some hexagon instabilities in long rollouts
One possible way to change that is to do the additative noise as in the original MeshGraphNet
or mildly randomize graph connectivity in encoder, as a kind of edge Dropout



"""

from typing import Tuple

import einops
import h3
import numpy as np
import torch
from torch_geometric.data import Data

from graph_weather.models.layers.graph_net_block import MLP, GraphProcessor


class Encoder(torch.nn.Module):
    """Encoder graph model"""

    def __init__(
        self,
        lat_lons: list,
        resolution: int = 2,
        input_dim: int = 78,
        output_dim: int = 256,
        output_edge_dim: int = 256,
        hidden_dim_processor_node=256,
        hidden_dim_processor_edge=256,
        hidden_layers_processor_node=2,
        hidden_layers_processor_edge=2,
        mlp_norm_type="LayerNorm",
        use_checkpointing: bool = False,
    ):
        """
        Encode the lat/lon data inot the isohedron graph

        Args:
            lat_lons: List of (lat,lon) points
            resolution: H3 resolution level
            input_dim: Input node dimension
            output_dim: Output node dimension
            output_edge_dim: Edge dimension
            hidden_dim_processor_node: Hidden dimension of the node processors
            hidden_dim_processor_edge: Hidden dimension of the edge processors
            hidden_layers_processor_node: Number of hidden layers in the node processors
            hidden_layers_processor_edge: Number of hidden layers in the edge processors
            mlp_norm_type: Type of norm for the MLPs
                one of 'LayerNorm', 'GraphNorm', 'InstanceNorm', 'BatchNorm', 'MessageNorm', or None
            use_checkpointing: Whether to use gradient checkpointing to use less memory
        """
        super().__init__()
        self.use_checkpointing = use_checkpointing
        self.output_dim = output_dim
        self.num_latlons = len(lat_lons)
        self.base_h3_grid = sorted(list(h3.uncompact(h3.get_res0_indexes(), resolution)))
        self.base_h3_map = {h_i: i for i, h_i in enumerate(self.base_h3_grid)}
        self.h3_grid = [h3.geo_to_h3(lat, lon, resolution) for lat, lon in lat_lons]
        self.h3_mapping = {}
        h_index = len(self.base_h3_grid)
        for h in self.base_h3_grid:
            if h not in self.h3_mapping:
                h_index -= 1
                self.h3_mapping[h] = h_index + self.num_latlons
        # Now have the h3 grid mapping, the bipartite graph of edges connecting lat/lon to h3 nodes
        # Should have vertical and horizontal difference
        self.h3_distances = []
        for idx, h3_point in enumerate(self.h3_grid):
            lat_lon = lat_lons[idx]
            distance = h3.point_dist(lat_lon, h3.h3_to_geo(h3_point), unit="rads")
            self.h3_distances.append([np.sin(distance), np.cos(distance)])
        self.h3_distances = torch.tensor(self.h3_distances, dtype=torch.float)
        # Compress to between 0 and 1

        # Build the default graph
        # lat_nodes = torch.zeros((len(lat_lons_heights), input_dim), dtype=torch.float)
        # h3_nodes = torch.zeros((h3.num_hexagons(resolution), output_dim), dtype=torch.float)
        # Get connections between lat nodes and h3 nodes
        edge_sources = []
        edge_targets = []
        for node_index, lat_node in enumerate(self.h3_grid):
            edge_sources.append(node_index)
            edge_targets.append(self.h3_mapping[lat_node])
        edge_index = torch.tensor([edge_sources, edge_targets], dtype=torch.long)

        # Use homogenous graph to make it easier
        self.graph = Data(edge_index=edge_index, edge_attr=self.h3_distances)

        self.latent_graph = self.create_latent_graph()

        # Extra starting ones for appending to inputs, could 'learn' good starting points
        self.h3_nodes = torch.nn.Parameter(
            torch.zeros((h3.num_hexagons(resolution), input_dim), dtype=torch.float)
        )
        # Output graph

        self.node_encoder = MLP(
            input_dim,
            output_dim,
            hidden_dim_processor_node,
            hidden_layers_processor_node,
            mlp_norm_type,
            self.use_checkpointing,
        )
        self.edge_encoder = MLP(
            2,
            output_edge_dim,
            hidden_dim_processor_edge,
            hidden_layers_processor_edge,
            mlp_norm_type,
            self.use_checkpointing,
        )
        self.latent_edge_encoder = MLP(
            2,
            output_edge_dim,
            hidden_dim_processor_edge,
            hidden_layers_processor_edge,
            mlp_norm_type,
            self.use_checkpointing,
        )
        self.graph_processor = GraphProcessor(
            1,
            output_dim,
            output_edge_dim,
            hidden_dim_processor_node,
            hidden_dim_processor_edge,
            hidden_layers_processor_node,
            hidden_layers_processor_edge,
            mlp_norm_type,
        )

    def forward(self, features: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Adds features to the encoding graph

        Args:
            features: Array of features in same order as lat_lon

        Returns:
            Torch tensors of node features, latent graph edge index, and latent edge attributes
        """
        batch_size = features.shape[0]
        self.h3_nodes = self.h3_nodes.to(features.device)
        self.graph = self.graph.to(features.device)
        self.latent_graph = self.latent_graph.to(features.device)
        features = torch.cat(
            [features, einops.repeat(self.h3_nodes, "n f -> b n f", b=batch_size)], dim=1
        )
        # Cat with the h3 nodes to have correct amount of nodes, and in right order
        features = einops.rearrange(features, "b n f -> (b n) f")
        out = self.node_encoder(features)  # Encode to 256 from 78
        edge_attr = self.edge_encoder(self.graph.edge_attr)  # Update attributes based on distance
        # Copy attributes batch times
        edge_attr = einops.repeat(edge_attr, "e f -> (repeat e) f", repeat=batch_size)
        # Expand edge index correct number of times while adding the proper number to the edge index
        edge_index = torch.cat(
            [
                self.graph.edge_index + i * torch.max(self.graph.edge_index) + i
                for i in range(batch_size)
            ],
            dim=1,
        )
        out, _ = self.graph_processor(out, edge_index, edge_attr)  # Message Passing
        # Remove the extra nodes (lat/lon) from the output
        out = einops.rearrange(out, "(b n) f -> b n f", b=batch_size)
        _, out = torch.split(out, [self.num_latlons, self.h3_nodes.shape[0]], dim=1)
        out = einops.rearrange(out, "b n f -> (b n) f")
        return (
            out,
            torch.cat(
                [
                    self.latent_graph.edge_index + i * torch.max(self.latent_graph.edge_index) + i
                    for i in range(batch_size)
                ],
                dim=1,
            ),
            self.latent_edge_encoder(
                einops.repeat(self.latent_graph.edge_attr, "e f -> (repeat e) f", repeat=batch_size)
            ),
        )  # New graph

    def create_latent_graph(self) -> Data:
        """
        Copies over and generates a Data object for the processor to use

        Returns:
            The connectivity and edge attributes for the latent graph
        """
        # Get connectivity of the graph
        edge_sources = []
        edge_targets = []
        edge_attrs = []
        for h3_index in self.base_h3_grid:
            h_points = h3.k_ring(h3_index, 1)
            for h in h_points:  # Already includes itself
                distance = h3.point_dist(h3.h3_to_geo(h3_index), h3.h3_to_geo(h), unit="rads")
                edge_attrs.append([np.sin(distance), np.cos(distance)])
                edge_sources.append(self.base_h3_map[h3_index])
                edge_targets.append(self.base_h3_map[h])
        edge_index = torch.tensor([edge_sources, edge_targets], dtype=torch.long)
        edge_attrs = torch.tensor(edge_attrs, dtype=torch.float)
        # Use heterogeneous graph as input and output dims are not same for the encoder
        # Because uniform grid now, don't need edge attributes as they are all the same
        return Data(edge_index=edge_index, edge_attr=edge_attrs)



"""
Training script for GenCast.
"""

import os

os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
os.environ["CUDA_VISIBLE_DEVICES"] = "0,3"

import lightning as L  # noqa: E402
import matplotlib.pyplot as plt  # noqa: E402
import numpy as np  # noqa: E402
import torch  # noqa: E402
from lightning.pytorch.callbacks import Callback, LearningRateMonitor, ModelCheckpoint  # noqa: E402
from lightning.pytorch.loggers import WandbLogger  # noqa: E402
from torch.utils.data import DataLoader  # noqa: E402

from graph_weather.data.gencast_dataloader import GenCastDataset  # noqa: E402
from graph_weather.models.gencast import Denoiser, Sampler, WeightedMSELoss  # noqa: E402

torch.set_float32_matmul_precision("high")

############################################## SETTINGS ############################################

# training settings
NUM_EPOCHS = 20
NUM_DEVICES = 2
NUM_ACC_GRAD = 1
INITIAL_LR = 1e-3
BATCH_SIZE = 16  # true batch size: BATCH_SIZE*NUM_DEVICES*NUM_ACC_GRAD
WARMUP = 1000

# dataloader setting
NUM_WORKERS = 8
PREFETCH_FACTOR = 3
PERSISTENT_WORKERS = True

# model configs
CHECKPOINT_PATH = "checkpoints/epoch=3-step=10776.ckpt"
CFG = {
    "hidden_dims": [512, 512],
    "num_blocks": 16,
    "num_heads": 4,
    "splits": 4,
    "num_hops": 8,
    "sparse": True,
    "use_edges_features": False,
    "scale_factor": 1.0,
}

# dataset configs
atmospheric_features = [
    "geopotential",
    "specific_humidity",
    "temperature",
    "u_component_of_wind",
    "v_component_of_wind",
    "vertical_velocity",
]
single_features = [
    "2m_temperature",
    "10m_u_component_of_wind",
    "10m_v_component_of_wind",
    "mean_sea_level_pressure",
    # "sea_surface_temperature",
    "total_precipitation_12hr",
]
static_features = [
    "geopotential_at_surface",
    "land_sea_mask",
]

OBS_PATH = "dataset.zarr"
# OBS_PATH = "gs://weatherbench2/datasets/era5/1959-2022-6h-128x64_equiangular_conservative.zarr"
# OBS_PATH = 'gs://weatherbench2/datasets/era5/1959-2022-6h-1440x721.zarr'
# OBS_PATH = 'gs://weatherbench2/datasets/era5/1959-2022-6h-512x256_equiangular_conservative.zarr'

#################################################################################################


class CosineWarmupScheduler(torch.optim.lr_scheduler._LRScheduler):
    """Cosine Scheduler with Warmup"""

    def __init__(self, optimizer, warmup, max_iters):
        """Initialize the scheduler"""
        self.warmup = warmup
        self.max_num_iters = max_iters
        super().__init__(optimizer)

    def get_lr(self):
        """Return the learning rates"""
        lr_factor = self.get_lr_factor(epoch=self.last_epoch)
        return [base_lr * lr_factor for base_lr in self.base_lrs]

    def get_lr_factor(self, epoch):
        """Return the scaling factor for the learning rate at a given iteration"""
        lr_factor = 0.5 * (1 + np.cos(np.pi * epoch / self.max_num_iters))
        if epoch <= self.warmup:
            lr_factor *= epoch * 1.0 / self.warmup
        return lr_factor


class LitModel(L.LightningModule):
    """Lightning wrapper for Gencast"""

    def __init__(
        self,
        warmup,
        learning_rate,
        cosine_t_max,
        pressure_levels,
        grid_lon,
        grid_lat,
        input_features_dim,
        output_features_dim,
        hidden_dims,
        num_blocks,
        num_heads,
        splits,
        num_hops,
        sparse,
        use_edges_features,
        scale_factor=1.0,
    ):
        """Initialize the module"""
        super().__init__()

        self.model = Denoiser(
            grid_lon=grid_lon,
            grid_lat=grid_lat,
            input_features_dim=input_features_dim,
            output_features_dim=output_features_dim,
            hidden_dims=hidden_dims,
            num_blocks=num_blocks,
            num_heads=num_heads,
            splits=splits,
            num_hops=num_hops,
            device=self.device,
            sparse=sparse,
            use_edges_features=use_edges_features,
            scale_factor=scale_factor,
        )

        self.criterion = WeightedMSELoss(
            grid_lat=torch.tensor(grid_lat).to(self.device),
            pressure_levels=torch.tensor(pressure_levels).to(self.device),
            num_atmospheric_features=len(atmospheric_features),
            single_features_weights=torch.tensor([1.0, 0.1, 0.1, 0.1, 0.1]).to(self.device),
        )

        self.learning_rate = learning_rate
        self.cosine_t_max = cosine_t_max
        self.warmup = warmup

    def forward(self, corrupted_targets, prev_inputs, noise_levels):
        """Compute forward pass"""
        return self.model(corrupted_targets, prev_inputs, noise_levels)

    def training_step(self, batch):
        """Single training step"""
        corrupted_targets, prev_inputs, noise_levels, target_residuals = batch

        preds = self.model(
            corrupted_targets=corrupted_targets,
            prev_inputs=prev_inputs,
            noise_levels=noise_levels,
        )
        loss = self.criterion(preds, noise_levels, target_residuals)
        self.log("loss", loss)
        return loss

    def configure_optimizers(self):
        """Initialize the optimizer"""
        opt = torch.optim.AdamW(
            self.parameters(), lr=self.learning_rate, weight_decay=0.1, betas=(0.9, 0.95)
        )
        # sch = torch.optim.lr_scheduler.CosineAnnealingLR(opt, T_max=self.cosine_t_max)
        sch = CosineWarmupScheduler(opt, warmup=self.warmup, max_iters=self.cosine_t_max)
        return {
            "optimizer": opt,
            "lr_scheduler": {
                "scheduler": sch,
                "monitor": "train_loss",
                "interval": "step",  # step means "batch" here, default: epoch
                "frequency": 1,  # default
            },
        }

    def plot_sample(self, prev_inputs, target_residuals):
        """Plot 2m_temperature and geopotential"""
        prev_inputs = prev_inputs[:1, :, :, :]
        target = target_residuals[:1, :, :, :]
        sampler = Sampler()
        preds = sampler.sample(self.model, prev_inputs)

        fig1, ax = plt.subplots(2)
        ax[0].imshow(preds[0, :, :, 78].T.cpu(), origin="lower", cmap="RdBu", vmin=-5, vmax=5)
        ax[0].set_xticks([])
        ax[0].set_yticks([])
        ax[0].set_title("Diffusion sampling prediction")

        ax[1].imshow(target[0, :, :, 78].T.cpu(), origin="lower", cmap="RdBu", vmin=-5, vmax=5)
        ax[1].set_xticks([])
        ax[1].set_yticks([])
        ax[1].set_title("Ground truth")

        fig2, ax = plt.subplots(2)
        ax[0].imshow(preds[0, :, :, 12].T.cpu(), origin="lower", cmap="RdBu", vmin=-5, vmax=5)
        ax[0].set_xticks([])
        ax[0].set_yticks([])
        ax[0].set_title("Diffusion sampling prediction")

        ax[1].imshow(target[0, :, :, 12].T.cpu(), origin="lower", cmap="RdBu", vmin=-5, vmax=5)
        ax[1].set_xticks([])
        ax[1].set_yticks([])
        ax[1].set_title("Ground truth")

        return fig1, fig2


class SamplingCallback(Callback):
    """Callback for sampling when a new epoch starts"""

    def __init__(self, data):
        """Initialize the callback"""
        _, prev_inputs, _, target_residuals = data
        self.prev_inputs = torch.tensor(prev_inputs).unsqueeze(0)
        self.target_residuals = torch.tensor(target_residuals).unsqueeze(0)

    def on_train_epoch_start(self, trainer, pl_module):
        """Sample and log predictions"""
        print("Epoch is starting")
        fig1, fig2 = pl_module.plot_sample(
            self.prev_inputs.to(pl_module.device), self.target_residuals.to(pl_module.device)
        )
        trainer.logger.log_image(
            key="samples", images=[fig1, fig2], caption=["2m_temperature", "geopotential"]
        )
        print("Uploaded samples")


if __name__ == "__main__":
    # define dataloader
    dataset = GenCastDataset(
        obs_path=OBS_PATH,
        atmospheric_features=atmospheric_features,
        single_features=single_features,
        static_features=static_features,
        max_year=2018,
        time_step=2,
    )

    dataloader = DataLoader(
        dataset,
        shuffle=True,
        batch_size=BATCH_SIZE,
        num_workers=NUM_WORKERS,
        prefetch_factor=PREFETCH_FACTOR,
        persistent_workers=PERSISTENT_WORKERS,
        multiprocessing_context="forkserver",
    )

    # define/resume model
    num_steps = NUM_EPOCHS * len(dataloader) // (NUM_DEVICES * NUM_ACC_GRAD)
    initial_lr = INITIAL_LR

    denoiser = LitModel.load_from_checkpoint(
        checkpoint_path=CHECKPOINT_PATH,
        warmup=WARMUP,
        learning_rate=initial_lr,
        cosine_t_max=num_steps,
        pressure_levels=dataset.pressure_levels,
        grid_lon=dataset.grid_lon,
        grid_lat=dataset.grid_lat,
        input_features_dim=dataset.input_features_dim,
        output_features_dim=dataset.output_features_dim,
        **CFG,
    )
    # denoiser = torch.compile(denoiser)

    # define trainer
    wandb_logger = WandbLogger(project="gencast")
    checkpoint_callback = ModelCheckpoint(dirpath="checkpoints/")
    lr_monitor = LearningRateMonitor(logging_interval="step")

    trainer = L.Trainer(
        accumulate_grad_batches=NUM_ACC_GRAD,
        accelerator="gpu",
        devices=NUM_DEVICES,
        strategy="DDP",
        # precision=16,
        max_epochs=NUM_EPOCHS,
        logger=wandb_logger,
        callbacks=[checkpoint_callback, SamplingCallback(data=dataset[0]), lr_monitor],
        log_every_n_steps=1,
    )

    # start training
    print("Starting training")
    trainer.fit(model=denoiser, train_dataloaders=dataloader)



"""
The dataloader for GenCast.

It has to:
- load, normalize and concatenate (across the channel dimension) the input timesteps 0 and 1.
- load and normalize the residual between timesteps 2 and 1.
- sample a noise level.
- corrupt the residual with noise generated at the given noise level.
"""

import warnings

import einops
import numpy as np
import xarray as xr
from torch.utils.data import Dataset

from graph_weather.data import const
from graph_weather.models.gencast.utils.noise import generate_isotropic_noise, sample_noise_level


class GenCastDataset(Dataset):
    """
    Dataset class for GenCast training data.
    """

    def __init__(
        self,
        obs_path: str,
        atmospheric_features: list[str],
        single_features: list[str],
        static_features: list[str],
        max_year: int = 2018,
        time_step: int = 2,
    ):
        """
        Initialize the GenCast dataset object.

        Args:
            obs_path: dataset path.
            atmospheric_features: list of features depending on pressure levels.
            single_features: list of features not depending on pressure levels.
            static_features: list of features not depending on time.
            max_year: max year to include in training set. Defaults to 2018.
            time_step: time step between predictions.
                        E.g. 12h steps correspond to time_step = 2 in a 6h dataset. Defaults to 2.
        """
        super().__init__()
        self.data = xr.open_zarr(obs_path, chunks={})
        self.max_year = max_year

        self.grid_lon = self.data["longitude"].values
        self.grid_lat = self.data["latitude"].values
        self.num_lon = len(self.grid_lon)
        self.num_lat = len(self.grid_lat)
        self.num_vars = len(self.data.keys())
        self.pressure_levels = np.array(self.data["level"].values).astype(
            np.float32
        )  # Need them for loss weighting
        self.output_features_dim = len(atmospheric_features) * len(self.pressure_levels) + len(
            single_features
        )
        self.input_features_dim = self.output_features_dim + len(static_features) + 4

        self.time_step = time_step  # e.g. 12h steps correspond to time_step = 2 in a 6h dataset

        self.atmospheric_features = atmospheric_features
        self.single_features = single_features
        self.static_features = static_features
        # Lat and long will be added by the model itself in the graph

        self.means, self.stds, self.diff_means, self.diff_stds = self._init_means_and_stds()

        # check if fast isotropic noise generation is possible
        if (self.num_lon == 2 * self.num_lat) or (self.num_lon == 2 * (self.num_lat - 1)):
            self.use_isotropic_noise = True
        else:
            self.use_isotropic_noise = False
            warnings.warn(
                "Isotropic noise requires grid's shape to be 2N x N or 2N x (N+1): "
                f"got {self.num_lon} x {self.num_lat}: falling back to flat normal random noise"
            )

    def _init_means_and_stds(self):
        means = []
        stds = []
        diff_means = []
        diff_stds = []

        for var in self.atmospheric_features:
            means.extend(const.ERA5_MEANS[var])
            stds.extend(const.ERA5_STD[var])
            diff_means.extend(const.ERA5_DIFF_MEAN[var])
            diff_stds.extend(const.ERA5_DIFF_STD[var])

        for var in self.single_features:
            means.append(const.ERA5_MEANS[var])
            stds.append(const.ERA5_STD[var])
            diff_means.append(const.ERA5_DIFF_MEAN[var])
            diff_stds.append(const.ERA5_DIFF_STD[var])

        for var in self.static_features:
            means.append(const.ERA5_MEANS[var])
            stds.append(const.ERA5_STD[var])

        return (
            np.array(means).astype(np.float32),
            np.array(stds).astype(np.float32),
            np.array(diff_means).astype(np.float32),
            np.array(diff_stds).astype(np.float32),
        )

    def _normalize(self, data, means, stds):
        return (data - means) / (stds + 0.0001)

    def _sin_cos_emb(self, x):
        return np.sin(2 * np.pi * x), np.cos(2 * np.pi * x)

    def _generate_clock_features(self, ds):
        # Compute sin/cos embedding for day of the year
        day_of_year = ds.time.dt.dayofyear.values
        day_of_year_grid = einops.repeat(
            day_of_year, "t -> t lon lat", lon=self.num_lon, lat=self.num_lat
        )
        sin_day_of_year, cos_day_of_year = self._sin_cos_emb(day_of_year_grid / 365.0)

        # Compute sin/cos embedding for local mean time
        hour_of_day = ds.time.dt.hour.values
        hour_of_day_grid = einops.repeat(
            hour_of_day, "t -> t lon lat", lon=self.num_lon, lat=self.num_lat
        )
        local_mean_time = hour_of_day_grid + ds["longitude"].values[None, :, None] * 4 / 60.0
        sin_local_mean_time, cos_local_mean_time = self._sin_cos_emb(local_mean_time / 24.0)

        # Stack clock features
        clock_input_data = np.stack(
            [sin_day_of_year, cos_day_of_year, sin_local_mean_time, cos_local_mean_time], axis=-1
        ).astype(np.float32)

        return clock_input_data

    def __len__(self):
        return sum(self.data["time.year"].values < self.max_year) - 2 * self.time_step

    def __getitem__(self, item):
        ds_inputs = self.data.isel(time=[item, item + self.time_step])
        ds_target = self.data.isel(time=item + 2 * self.time_step)

        # Load inputs data
        ds_inputs_atm = (
            ds_inputs[self.atmospheric_features]
            .to_array()
            .transpose("time", "longitude", "latitude", "level", "variable")
            .values
        )
        ds_inputs_atm = einops.rearrange(ds_inputs_atm, "t lon lat lev var -> t lon lat (var lev)")
        ds_inputs_single = (
            ds_inputs[self.single_features]
            .to_array()
            .transpose("time", "longitude", "latitude", "variable")
            .values
        )
        ds_inputs_static = (
            ds_inputs[self.static_features]
            .to_array()
            .transpose("longitude", "latitude", "variable")
            .values
        )
        ds_inputs_static = np.stack([ds_inputs_static] * 2, axis=0)

        raw_inputs = np.concatenate([ds_inputs_atm, ds_inputs_single, ds_inputs_static], axis=-1)

        # Normalize inputs
        inputs_norm = self._normalize(raw_inputs, self.means, self.stds)

        # Add time features
        clock_features = self._generate_clock_features(ds_inputs)
        inputs = np.concatenate([inputs_norm, clock_features], axis=-1)

        # Concatenate timesteps
        inputs = np.concatenate([inputs[0, :, :, :], inputs[1, :, :, :]], axis=-1)
        prev_inputs = np.nan_to_num(inputs).astype(np.float32)

        # Load target data
        ds_target_atm = (
            ds_target[self.atmospheric_features]
            .to_array()
            .transpose("longitude", "latitude", "level", "variable")
            .values
        )
        ds_target_atm = einops.rearrange(ds_target_atm, "lon lat lev var -> lon lat (var lev)")
        ds_target_single = (
            ds_target[self.single_features]
            .to_array()
            .transpose("longitude", "latitude", "variable")
            .values
        )

        raw_targets = np.concatenate([ds_target_atm, ds_target_single], axis=-1)

        # Normalize target residuals
        raw_target_residuals = raw_targets - raw_inputs[1, :, :, : raw_targets.shape[-1]]
        target_residuals = self._normalize(raw_target_residuals, self.diff_means, self.diff_stds)
        target_residuals = np.nan_to_num(target_residuals).astype(np.float32)

        # Corrupt targets with noise
        noise_levels = np.array([sample_noise_level()]).astype(np.float32)
        noise = generate_isotropic_noise(
            num_lon=self.num_lon,
            num_lat=self.num_lat,
            num_samples=target_residuals.shape[-1],
            isotropic=self.use_isotropic_noise,
        )
        corrupted_targets = target_residuals + noise_levels * noise

        return (
            corrupted_targets,
            prev_inputs,
            noise_levels,
            target_residuals,
        )


class BatchedGenCastDataset(Dataset):
    """
    Dataset class for GenCast batched training data.

    This dataset object returns a full batch as a single sample, it may be faster.

    Args:
        obs_path: Dataset path.
        atmospheric_features: List of features dependent on pressure levels.
        single_features: List of features not dependent on pressure levels.
        static_features: List of features not dependent on time.
        max_year (optional): Max year to include in training set. Defaults to 2018.
        time_step (optional): Time step between predictions.
                    E.g. 12h steps correspond to time_step = 2 in a 6h dataset. Defaults to 2.
        batch_size (optional): Size of the batch. Defaults to 32.
    """

    def __init__(
        self,
        obs_path: str,
        atmospheric_features: list[str],
        single_features: list[str],
        static_features: list[str],
        max_year: int = 2018,
        time_step: int = 2,
        batch_size: int = 32,  # maybe make optional?
    ):
        """
        Initialize the GenCast dataset object.
        """
        super().__init__()
        self.data = xr.open_zarr(obs_path, chunks={})
        self.max_year = max_year

        self.grid_lon = self.data["longitude"].values
        self.grid_lat = self.data["latitude"].values
        self.num_lon = len(self.data["longitude"].values)
        self.num_lat = len(self.data["latitude"].values)
        self.num_vars = len(self.data.keys())
        self.pressure_levels = np.array(self.data["level"].values).astype(
            np.float32
        )  # Need them for loss weighting

        self.batch_size = batch_size
        self.time_step = time_step  # 12h steps correspond to time_step = 2 in a 6h dataset

        self.atmospheric_features = atmospheric_features
        self.single_features = single_features
        self.static_features = static_features
        # Lat and long will be added by the model itself in the graph

        self.clock_features = ["local_time_of_the_day", "elapsed_year_progress"]

        self.means, self.stds, self.diff_means, self.diff_stds = self._init_means_and_stds()

        # check if fast isotropic noise generation is possible
        if (self.num_lon == 2 * self.num_lat) or (self.num_lon == 2 * (self.num_lat - 1)):
            self.use_isotropic_noise = True
        else:
            self.use_isotropic_noise = False
            warnings.warn(
                "Isotropic noise requires grid's shape to be 2N x N or 2N x (N+1): "
                f"got {self.num_lon} x {self.num_lat}: falling back to flat normal random noise"
            )

    def _init_means_and_stds(self):
        means = []
        stds = []
        diff_means = []
        diff_stds = []

        for var in self.atmospheric_features:
            means.extend(const.ERA5_MEANS[var])
            stds.extend(const.ERA5_STD[var])
            diff_means.extend(const.ERA5_DIFF_MEAN[var])
            diff_stds.extend(const.ERA5_DIFF_STD[var])

        for var in self.single_features:
            means.append(const.ERA5_MEANS[var])
            stds.append(const.ERA5_STD[var])
            diff_means.append(const.ERA5_DIFF_MEAN[var])
            diff_stds.append(const.ERA5_DIFF_STD[var])

        for var in self.static_features:
            means.append(const.ERA5_MEANS[var])
            stds.append(const.ERA5_STD[var])

        return (
            np.array(means).astype(np.float32),
            np.array(stds).astype(np.float32),
            np.array(diff_means).astype(np.float32),
            np.array(diff_stds).astype(np.float32),
        )

    def _normalize(self, data, means, stds):
        return (data - means) / (stds + 0.0001)

    def _batchify_inputs(self, data):
        start_idx = []
        for i in range(self.batch_size):
            start_idx.append([i, i + self.time_step])
        return data[start_idx]

    def _batchify_diffs(self, data):
        prev_idx = []
        target_idx = []
        for i in range(self.batch_size):
            prev_idx.append(i + self.time_step)
            target_idx.append(i + 2 * self.time_step)
        return data[target_idx] - data[prev_idx]

    def _generate_clock_features(self, ds):
        day_of_year = ds.time.dt.dayofyear.values / 365.0
        sin_day_of_year = (
            np.sin(2 * np.pi * day_of_year)[:, None, None]
            * np.ones((self.num_lon, self.num_lat))[None, :, :]
        )
        cos_day_of_year = (
            np.cos(2 * np.pi * day_of_year)[:, None, None]
            * np.ones((self.num_lon, self.num_lat))[None, :, :]
        )

        local_mean_time = (
            np.ones((self.num_lon, self.num_lat))[None, :, :]
            * (ds.time.dt.hour.values[:, None, None])
            + ds["longitude"].values[None, :, None] * 4 / 60.0
        )
        sin_local_mean_time = np.sin(2 * np.pi * local_mean_time / 24.0)
        cos_local_mean_time = np.cos(2 * np.pi * local_mean_time / 24.0)

        clock_input_data = np.stack(
            [sin_day_of_year, cos_day_of_year, sin_local_mean_time, cos_local_mean_time], axis=-1
        ).astype(np.float32)
        return clock_input_data

    def __len__(self):
        return sum(self.data["time.year"].values < self.max_year) - (
            3 * self.time_step + self.batch_size - 2
        )

    def __getitem__(self, item):
        # Compute the starting and ending point of the batch.
        starting_point = self.batch_size * item
        ending_point = starting_point + 3 * self.time_step + self.batch_size - 2

        # Load data
        ds = self.data.isel(time=np.arange(starting_point, ending_point))
        ds_atm = (
            ds[self.atmospheric_features]
            .to_array()
            .transpose("time", "longitude", "latitude", "level", "variable")
            .values
        )
        ds_atm = einops.rearrange(ds_atm, "t lon lat lev var -> t lon lat (var lev)")
        ds_single = (
            ds[self.single_features]
            .to_array()
            .transpose("time", "longitude", "latitude", "variable")
            .values
        )
        ds_static = (
            ds[self.static_features]
            .to_array()
            .transpose("longitude", "latitude", "variable")
            .values
        )
        ds_static = np.stack([ds_static] * (ending_point - starting_point), axis=0)

        # Compute inputs
        raw_inputs = np.concatenate([ds_atm, ds_single, ds_static], axis=-1)
        batched_inputs = self._batchify_inputs(raw_inputs)
        batched_inputs_norm = self._normalize(batched_inputs, self.means, self.stds)

        # Add time features
        ds_clock = self._batchify_inputs(self._generate_clock_features(ds))
        inputs = np.concatenate([batched_inputs_norm, ds_clock], axis=-1)
        # Concatenate timesteps
        inputs = np.concatenate([inputs[:, 0, :, :, :], inputs[:, 1, :, :, :]], axis=-1)
        prev_inputs = np.nan_to_num(inputs).astype(np.float32)

        # Compute targets residuals
        raw_targets = np.concatenate([ds_atm, ds_single], axis=-1)
        batched_residuals = self._batchify_diffs(raw_targets)
        target_residuals = self._normalize(batched_residuals, self.diff_means, self.diff_stds)
        target_residuals = np.nan_to_num(target_residuals).astype(np.float32)

        # Corrupt targets with noise
        noise_levels = np.zeros((self.batch_size, 1), dtype=np.float32)
        corrupted_targets = np.zeros_like(target_residuals, dtype=np.float32)
        for b in range(self.batch_size):
            noise_level = sample_noise_level()
            noise = generate_isotropic_noise(
                num_lon=self.num_lon,
                num_lat=self.num_lat,
                num_samples=target_residuals.shape[-1],
                isotropic=self.use_isotropic_noise,
            )
            corrupted_targets[b] = target_residuals[b] + noise_level * noise
            noise_levels[b] = noise_level

        return (corrupted_targets, prev_inputs, noise_levels, target_residuals)



"""Encoder layer.

The encoder:
- embeds grid nodes, mesh nodes and g2m edges' features to the latent space.
- perform a single message-passing step using a classical interaction network.
- add a residual connection to the mesh and grid nodes.
"""

import torch

from graph_weather.models.gencast.layers.modules import MLP, InteractionNetwork


class Encoder(torch.nn.Module):
    """GenCast's encoder."""

    def __init__(
        self,
        grid_dim: int,
        mesh_dim: int,
        edge_dim: int,
        hidden_dims: list[int],
        activation_layer: torch.nn.Module = torch.nn.ReLU,
        use_layer_norm: bool = True,
        scale_factor: float = 1.0,
    ):
        """Initialize the Encoder.

        Args:
            grid_dim (int): dimension of grid nodes' features.
            mesh_dim (int): dimension of mesh nodes' features
            edge_dim (int): dimension of g2m edges' features
            hidden_dims (list[int]): hidden dimensions of internal MLPs.
            activation_layer (torch.nn.Module, optional): activation function of internal MLPs.
                Defaults to torch.nn.ReLU.
            use_layer_norm (bool, optional): if true add a LayerNorm at the end of each MLP.
                Defaults to True.
            scale_factor (float): the message of the interaction network between the grid and the
                the mesh is multiplied by the scale factor. Useful when fine-tuning a pretrained
                model to a higher resolution. Defaults to 1.
        """
        super().__init__()

        # All the MLPs in GenCast have same hidden and output dims. Hence, the embedding latent
        # dimension and the MLPs' output dimension are the same. Moreover, for simplicity, we will
        # ask the hidden dims just once for each MLP in a module: we don't need to specify them
        # individually as arguments, even if the MLPs could have different roles.
        self.latent_dim = hidden_dims[-1]

        # Embedders
        self.grid_mlp = MLP(
            input_dim=grid_dim,
            hidden_dims=hidden_dims,
            activation_layer=activation_layer,
            use_layer_norm=use_layer_norm,
            bias=True,
            activate_final=False,
        )

        self.mesh_mlp = MLP(
            input_dim=mesh_dim,
            hidden_dims=hidden_dims,
            activation_layer=activation_layer,
            use_layer_norm=use_layer_norm,
            bias=True,
            activate_final=False,
        )

        self.edges_mlp = MLP(
            input_dim=edge_dim,
            hidden_dims=hidden_dims,
            activation_layer=activation_layer,
            use_layer_norm=use_layer_norm,
            bias=True,
            activate_final=False,
        )

        # Message Passing
        self.gnn = InteractionNetwork(
            sender_dim=self.latent_dim,
            receiver_dim=self.latent_dim,
            edge_attr_dim=self.latent_dim,
            hidden_dims=hidden_dims,
            use_layer_norm=use_layer_norm,
            activation_layer=activation_layer,
            scale_factor=scale_factor,
        )

        # Final grid nodes update
        self.grid_mlp_final = MLP(
            input_dim=self.latent_dim,
            hidden_dims=hidden_dims,
            activation_layer=activation_layer,
            use_layer_norm=use_layer_norm,
            bias=True,
            activate_final=False,
        )

    def forward(
        self,
        input_grid_nodes: torch.Tensor,
        input_mesh_nodes: torch.Tensor,
        input_edge_attr: torch.Tensor,
        edge_index: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Forward pass.

        Args:
            input_grid_nodes (torch.Tensor): grid nodes' features.
            input_mesh_nodes (torch.Tensor): mesh nodes' features.
            input_edge_attr (torch.Tensor): grid2mesh edges' features.
            edge_index (torch.Tensor): edge index tensor.

        Returns:
            tuple[torch.Tensor, torch.Tensor]: output grid nodes, output mesh nodes.
        """

        # Embedding
        grid_emb = self.grid_mlp(input_grid_nodes)
        mesh_emb = self.mesh_mlp(input_mesh_nodes)
        edges_emb = self.edges_mlp(input_edge_attr)

        # Message-passing + residual connection
        latent_mesh_nodes = mesh_emb + self.gnn(
            x=(grid_emb, mesh_emb),
            edge_index=edge_index,
            edge_attr=edges_emb,
        )

        # Update grid nodes + residual connection
        latent_grid_nodes = grid_emb + self.grid_mlp_final(grid_emb)

        return latent_grid_nodes, latent_mesh_nodes



"""Modules"""

import math

import torch
import torch.nn as nn
from torch_geometric.nn import MessagePassing
from torch_geometric.nn.conv import TransformerConv


class MLP(nn.Module):
    """Classic multi-layer perceptron (MLP) module."""

    def __init__(
        self,
        input_dim: int,
        hidden_dims: list[int],
        activation_layer: nn.Module = nn.ReLU,
        use_layer_norm: bool = False,
        bias: bool = True,
        activate_final: bool = False,
    ):
        """Initialize MLP module.

        Args:
            input_dim (int): dimension of input.
            hidden_dims (List[int]): list of hidden linear layers dimensions.
            activation_layer (torch.nn.Module): activation
                function to use. Defaults to torch.nn.ReLU.
            use_layer_norm (bool, optional): if Ttrue apply LayerNorm to output. Defaults to False.
            bias (bool, optional): if true use bias in linear layers. Defaults to True.
            activate_final (bool, optional): whether to apply  the activation function to the final
                layer. Defaults to False.
        """
        super().__init__()

        # Initialize linear layers
        self.linears = nn.ModuleList()
        self.linears.append(nn.Linear(input_dim, hidden_dims[0], bias=bias))
        for i in range(0, len(hidden_dims) - 1):
            self.linears.append(nn.Linear(hidden_dims[i], hidden_dims[i + 1], bias=bias))

        # Initialize activation
        self.activation = activation_layer()

        # Initialize layer normalization
        self.norm_layer = None
        if use_layer_norm:
            self.norm_layer = nn.LayerNorm(hidden_dims[-1])

        self.activate_final = activate_final

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Apply MLP to input."""

        for linear in self.linears[:-1]:
            x = linear(x)
            x = self.activation(x)

        x = self.linears[-1](x)

        if self.activate_final:
            x = self.activation(x)

        if self.norm_layer is not None:
            x = self.norm_layer(x)
        return x


class InteractionNetwork(MessagePassing):
    """Single message-passing interaction network as described in GenCast.

    This network performs two steps:
    1) message-passing: e'_ij = MLP([e_ij,v_i,v_j])
    2) aggregation: v'_j = MLP([v_j, sum_i {e'_ij}])
    The underlying graph is a directed graph.

    Note:
        We don't need to update edges in GenCast, hence we skip it.
    """

    def __init__(
        self,
        sender_dim: int,
        receiver_dim: int,
        edge_attr_dim: int,
        hidden_dims: list[int],
        use_layer_norm: bool = False,
        activation_layer: nn.Module = nn.ReLU,
        scale_factor: float = 1.0,
    ):
        """Initialize the Interaction Network.

        Args:
            sender_dim (int): dimension of sender nodes' features.
            receiver_dim (int): dimension of receiver nodes' features.
            edge_attr_dim (int): dimension of the edge features.
            hidden_dims (list[int]): list of sizes of MLP's linear layers.
            use_layer_norm (bool): if true add layer normalization to MLP's last layer.
                Defaults to False.
            activation_layer (torch.nn.Module): activation function. Defaults to nn.ReLU.
            scale_factor (float): the message is multiplied by this value. Defaults to 1.0.
        """
        super().__init__(aggr="add", flow="source_to_target")
        self.mlp_edges = MLP(
            input_dim=sender_dim + receiver_dim + edge_attr_dim,
            hidden_dims=hidden_dims,
            activation_layer=activation_layer,
            use_layer_norm=use_layer_norm,
            bias=True,
            activate_final=False,
        )
        self.mlp_nodes = MLP(
            input_dim=receiver_dim + hidden_dims[-1],
            hidden_dims=hidden_dims,
            activation_layer=activation_layer,
            use_layer_norm=use_layer_norm,
            bias=True,
            activate_final=False,
        )
        self.scale_factor = scale_factor

    def message(self, x_i, x_j, edge_attr):
        """Message-passing step."""
        x = torch.cat((x_i, x_j, edge_attr), dim=-1)
        x = self.mlp_edges(x)
        return self.scale_factor * x

    def forward(
        self,
        x: tuple[torch.Tensor, torch.Tensor],
        edge_index: torch.Tensor,
        edge_attr: torch.Tensor,
    ) -> torch.Tensor:
        """
        Compute the output of the Interaction Network.

        This method processes the input node features and edge attributes through
        the network to produce the output node features.

        Args:
            x (tuple[torch.Tensor, torch.Tensor]): a tuple containing:
                - sender nodes' features (torch.Tensor): features of the sender nodes.
                - receiver nodes' features (torch.Tensor): features of the receiver nodes.
            edge_index (torch.Tensor): tensor containing edge indices, defining the
                connections between nodes.
            edge_attr (torch.Tensor): tensor containing edge features, representing
                the attributes of each edge.

        Returns:
            torch.Tensor: the resulting node features after applying the Interaction Network.
        """
        aggr = self.propagate(
            edge_index, x=x, edge_attr=edge_attr, size=(x[0].shape[0], x[1].shape[0])
        )
        out = self.mlp_nodes(torch.cat((x[1], aggr), dim=-1))
        return out


class FourierEmbedding(nn.Module):
    """Fourier embedding module."""

    def __init__(self, output_dim: int, num_frequencies: int, base_period: int):
        """Initialize the Fourier Embedder.

        Args:
            output_dim (int): output dimension.
            num_frequencies (int): number of frequencies for sin/cos embedding.
            base_period (int): max period for sin/cos embedding.
        """
        super().__init__()
        self.mlp = torch.nn.Sequential(
            nn.Linear(2 * num_frequencies, output_dim, bias=True),
            nn.SiLU(),
            nn.Linear(output_dim, output_dim, bias=True),
        )
        self.num_frequencies = num_frequencies
        self.base_period = base_period

    def fourier_features(self, t):
        """
        Create sinusoidal embeddings.
        """
        freqs = torch.exp(
            -math.log(self.base_period)
            * torch.arange(start=0, end=self.num_frequencies, dtype=torch.float32)
            / self.num_frequencies
        ).to(device=t.device)
        args = t * freqs[None, :]
        embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1)
        return embedding

    def forward(self, t):
        """Apply fourier features and mlp to the input."""
        t = self.fourier_features(t)
        t = self.mlp(t)
        return t


class ConditionalLayerNorm(nn.Module):
    """Conditional Layer Normalization.

    This module is a variant of Layer Normalization: an elementwise affine transformation is applied
    to the output of the LayerNorm, with parameters computed as Linears of some conditioning params.
    """

    def __init__(
        self,
        conditioning_dim: int,
        features_dim: int,
    ):
        """Initialize Conditional Layer Normalization module.

        Args:
            conditioning_dim (int): dimension of the conditioning parameter.
            features_dim (int): dimension of the input features.
        """
        super().__init__()

        # Initialize linear layers
        self.linear_scale = nn.Linear(conditioning_dim, features_dim, bias=True)
        self.linear_bias = nn.Linear(conditioning_dim, features_dim, bias=True)

        # Initialize the LayerNorm
        self.norm = nn.LayerNorm(features_dim, eps=1e-05, elementwise_affine=False)

    def forward(self, x: torch.Tensor, cond_param: torch.Tensor) -> torch.Tensor:
        """Apply ConditionalLayerNorm to input.

        Input and conditioning parameter must have same batch size.

        Args:
            x (torch.Tensor): input.
            cond_param (torch.Tensor): conditioning parameter.
        """

        # check shapes.
        if not x.shape[0] == cond_param.shape[0]:
            raise ValueError(
                "Expected same batch dimension for the input and the conditioning "
                f"parameter, got {x.shape[0]} and {cond_param.shape[0]}"
            )

        # compute parameters and normalize.
        scale = self.linear_scale(cond_param)
        bias = self.linear_bias(cond_param)
        x_norm = self.norm(x)

        assert scale.shape[1] == x_norm.shape[1]

        # apply elementwise affine transformation.
        out = scale * x_norm + bias
        return out


class CondTransformerBlock(nn.Module):
    """
    A single transformer block for graph neural networks.

    This module implements a transformer block tailored for graph structures, following the
    methodology outlined in the paper "Masked Label Prediction: Unified Message Passing Model
    for Semi-Supervised Classification." The implementation aligns with the principles described
    in this paper, providing an adaptation of the transformer's attention mechanism to
    graph data.
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_heads: int,
        conditioning_dim: int | None = None,
        edges_dim: int | None = None,
        concat: bool = True,
        beta: bool = True,
        activation_layer: torch.nn.Module | None = nn.ReLU,
    ):
        """Initialize Conditional Layer Normalization module.

        Args:
            input_dim (int): dimension of the input features.
            output_dim (int): dimension of the output features.
            edges_dim (int): dimension of the edge features.
            num_heads (int): number of heads for multi-head attention.
            conditioning_dim (int, optional): dimension of the conditioning parameter. If None the
                layer normalization will not be applied.
            edges_dim (int, optional): dimension of the edges features. If None edges features will
                not be used inside TransformerConv.
            concat (bool): if true concatenate the outputs of each head, otherwise average them.
                Defaults to True.
            beta (bool): if true apply the beta weighting described in the paper. Defauls to True.
            activation_layer (torch.nn.Module, optional): activation function applied before
                returning the output. If None skip the activation function. Defaults to nn.ReLU.
        """
        super().__init__()

        # Initialize layers
        self.transformer_conv = TransformerConv(
            in_channels=input_dim,
            out_channels=output_dim,
            heads=num_heads,
            concat=concat,
            beta=beta,
            edge_dim=edges_dim,
        )

        self.activation = activation_layer() if activation_layer is not None else None

        if conditioning_dim is not None:
            final_dim = num_heads * output_dim if concat else output_dim
            self.cond_norm = ConditionalLayerNorm(
                conditioning_dim=conditioning_dim, features_dim=final_dim
            )
        else:
            self.cond_norm = None

    def forward(
        self,
        x: torch.Tensor,
        edge_index: torch.Tensor,
        edge_attr: torch.Tensor | None = None,
        cond_param: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """Apply CondTransformerBlock to input.

        Input and conditioning parameter must have same batch size.

        Args:
            x (torch.Tensor): tensor containing nodes features.
            edge_index (torch.Tensor): edge index tensor.
            edge_attr (torch.Tensor, optional): tensor containing edges features.
            cond_param (torch.Tensor, optional): conditioning parameter.

        """
        x = self.transformer_conv(x=x, edge_index=edge_index, edge_attr=edge_attr)

        if self.cond_norm is not None:
            x = self.cond_norm(x, cond_param)

        if self.activation is not None:
            x = self.activation(x)

        return x



"""Models"""

from .fengwu_ghr.layers import ImageMetaModel, MetaModel, WrapperImageModel, WrapperMetaModel
from .layers.assimilator_decoder import AssimilatorDecoder
from .layers.assimilator_encoder import AssimilatorEncoder
from .layers.decoder import Decoder
from .layers.encoder import Encoder
from .layers.processor import Processor



"""

The dataloader has to do a few things for the model to work correctly

1. Load the land-0sea mask, orography dataset, regridded from 0.1 to the
correct resolution
2. Calculate the top-of-atmosphere solar radiation for each location at
fcurrent time and 10 other
 times +- 12 hours
3. Add day-of-year, sin(lat), cos(lat), sin(lon), cos(lon) as well
3. Batch data as either in geometric batches, or more normally
4. Rescale between 0 and 1, but don't normalize

"""

import const
import numpy as np
import pandas as pd
import xarray as xr
from pysolar.util import extraterrestrial_irrad
from torch.utils.data import Dataset


class AnalysisDataset(Dataset):
    """
    Dataset class for analysis data.

    Args:
        filepaths: List of file paths.
        invariant_path: Path to the invariant file.
        mean: Mean value.
        std  Standard deviation value.
        coarsen : Coarsening factor. Defaults to 8.

    Methods:
        __init__: Initialize the AnalysisDataset object.
        __len__: Get the length of the dataset.
        __getitem__: Get an item from the dataset.
    """

    def __init__(self, filepaths, invariant_path, mean, std, coarsen: int = 8):
        """
        Initialize the AnalysisDataset object.
        """
        super().__init__()
        self.filepaths = sorted(filepaths)
        self.invariant_path = invariant_path
        self.coarsen = coarsen
        self.mean = mean
        self.std = std

    def __len__(self):
        return len(self.filepaths) - 1

    def __getitem__(self, item):
        if self.coarsen <= 1:  # Don't coarsen, so don't even call it
            start = xr.open_zarr(self.filepaths[item], consolidated=True)
            end = xr.open_zarr(self.filepaths[item + 1], consolidated=True)
        else:
            start = (
                xr.open_zarr(self.filepaths[item], consolidated=True)
                .coarsen(latitude=self.coarsen, boundary="pad")
                .mean()
                .coarsen(longitude=self.coarsen)
                .mean()
            )
            end = (
                xr.open_zarr(self.filepaths[item + 1], consolidated=True)
                .coarsen(latitude=self.coarsen, boundary="pad")
                .mean()
                .coarsen(longitude=self.coarsen)
                .mean()
            )

        # Land-sea mask data, resampled to the same as the physical variables
        landsea = (
            xr.open_zarr(self.invariant_path, consolidated=True)
            .interp(latitude=start.latitude.values)
            .interp(longitude=start.longitude.values)
        )
        # Calculate sin,cos, day of year, solar irradiance here before stacking
        landsea = np.stack(
            [
                (landsea[f"{var}"].values - const.LANDSEA_MEAN[var]) / const.LANDSEA_STD[var]
                for var in landsea.data_vars
                if not np.isnan(landsea[f"{var}"].values).any()
            ],
            axis=-1,
        )
        landsea = landsea.T.reshape((-1, landsea.shape[-1]))
        lat_lons = np.array(np.meshgrid(start.latitude.values, start.longitude.values)).T.reshape(
            (-1, 2)
        )
        sin_lat_lons = np.sin(lat_lons)
        cos_lat_lons = np.cos(lat_lons)
        date = start.time.dt.values
        day_of_year = start.time.dayofyear.values / 365.0
        np.sin(day_of_year)
        np.cos(day_of_year)
        solar_times = [np.array([extraterrestrial_irrad(date, lat, lon) for lat, lon in lat_lons])]
        for when in pd.date_range(
            date - pd.Timedelta("12 hours"), date + pd.Timedelta("12 hours"), freq="1H"
        ):
            solar_times.append(
                np.array([extraterrestrial_irrad(when, lat, lon) for lat, lon in lat_lons])
            )
        solar_times = np.array(solar_times)

        # End time solar radiation too
        end_date = end.time.dt.values
        end_solar_times = [
            np.array([extraterrestrial_irrad(end_date, lat, lon) for lat, lon in lat_lons])
        ]
        for when in pd.date_range(
            end_date - pd.Timedelta("12 hours"), end_date + pd.Timedelta("12 hours"), freq="1H"
        ):
            end_solar_times.append(
                np.array([extraterrestrial_irrad(when, lat, lon) for lat, lon in lat_lons])
            )
        end_solar_times = np.array(solar_times)

        # Normalize to between -1 and 1
        solar_times -= const.SOLAR_MEAN
        solar_times /= const.SOLAR_STD
        end_solar_times -= const.SOLAR_MEAN
        end_solar_times /= const.SOLAR_STD

        # Stack the data into a large data cube
        input_data = np.stack(
            [
                start[f"{var}"].values
                for var in start.data_vars
                if not np.isnan(start[f"{var}"].values).any()
            ],
            axis=-1,
        )
        # TODO Combine with above? And include sin/cos of day of year
        input_data = np.concatenate(
            [
                input_data.T.reshape((-1, input_data.shape[-1])),
                sin_lat_lons,
                cos_lat_lons,
                solar_times,
                landsea,
            ],
            axis=-1,
        )
        # Not want to predict non-physics variables -> Output only the data variables?
        # Would be simpler, and just add in the new ones each time

        output_data = np.stack(
            [
                end[f"{var}"].values
                for var in end.data_vars
                if not np.isnan(end[f"{var}"].values).any()
            ],
            axis=-1,
        )

        output_data = np.concatenate(
            [
                output_data.T.reshape((-1, output_data.shape[-1])),
                sin_lat_lons,
                cos_lat_lons,
                end_solar_times,
                landsea,
            ],
            axis=-1,
        )
        # Stick with Numpy, don't tensor it, as just going from 0 to 1

        # Normalize now
        return input_data, output_data


obs_data = xr.open_zarr(
    "/home/jacob/Development/prepbufr.gdas.20160101.t00z.nr.48h.raw.zarr", consolidated=True
)

# TODO Embedding? These should stay consistent across all of the inputs, so can just load the values
#  not the strings?
# Should only take in the quality markers, observations, reported observation time relative to start
# point
# Observation errors, and background values, lat/lon/height/speed of observing thing

print(obs_data)
print(obs_data.hdr_inst_typ.values)
print(obs_data.hdr_irpt_typ.values)
print(obs_data.obs_qty_table.values)
print(obs_data.hdr_prpt_typ.values)
print(obs_data.hdr_sid_table.values)
print(obs_data.hdr_typ_table.values)
print(obs_data.obs_desc.values)
print(obs_data.data_vars.keys())
exit()
analysis_data = xr.open_zarr(
    "/home/jacob/Development/gdas1.fnl0p25.2016010100.f00.zarr", consolidated=True
)
print(analysis_data)



# Copyright 2023 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Source: https://github.com/google-deepmind/graphcast.
"""Utilities for building models."""

from typing import Mapping, Optional, Tuple

import numpy as np
import xarray
from scipy.spatial import transform


def get_graph_spatial_features(
    *,
    node_lat: np.ndarray,
    node_lon: np.ndarray,
    senders: np.ndarray,
    receivers: np.ndarray,
    add_node_positions: bool,
    add_node_latitude: bool,
    add_node_longitude: bool,
    add_relative_positions: bool,
    relative_longitude_local_coordinates: bool,
    relative_latitude_local_coordinates: bool,
    sine_cosine_encoding: bool = False,
    encoding_num_freqs: int = 10,
    encoding_multiplicative_factor: float = 1.2,
) -> Tuple[np.ndarray, np.ndarray]:
    """Computes spatial features for the nodes.

    Args:
      node_lat: Latitudes in the [-90, 90] interval of shape [num_nodes]
      node_lon: Longitudes in the [0, 360] interval of shape [num_nodes]
      senders: Sender indices of shape [num_edges]
      receivers: Receiver indices of shape [num_edges]
      add_node_positions: Add unit norm absolute positions.
      add_node_latitude: Add a feature for latitude (cos(90 - lat))
          Note even if this is set to False, the model may be able to infer the
          longitude from relative features, unless
          `relative_latitude_local_coordinates` is also True, or if there is any
          bias on the relative edge sizes for different longitudes.
      add_node_longitude: Add features for longitude (cos(lon), sin(lon)).
          Note even if this is set to False, the model may be able to infer the
          longitude from relative features, unless
          `relative_longitude_local_coordinates` is also True, or if there is any
          bias on the relative edge sizes for different longitudes.
      add_relative_positions: Whether to relative positions in R3 to the edges.
      relative_longitude_local_coordinates: If True, relative positions are
          computed in a local space where the receiver is at 0 longitude.
      relative_latitude_local_coordinates: If True, relative positions are
          computed in a local space where the receiver is at 0 latitude.
      sine_cosine_encoding: If True, we will transform the node/edge features
          with sine and cosine functions, similar to NERF.
      encoding_num_freqs: frequency parameter
      encoding_multiplicative_factor: used for calculating the frequency.

    Returns:
      Arrays of shape: [num_nodes, num_features] and [num_edges, num_features].
      with node and edge features.

    """

    num_nodes = node_lat.shape[0]
    num_edges = senders.shape[0]
    dtype = node_lat.dtype
    node_phi, node_theta = lat_lon_deg_to_spherical(node_lat, node_lon)

    # Computing some node features.
    node_features = []
    if add_node_positions:
        # Already in [-1, 1.] range.
        node_features.extend(spherical_to_cartesian(node_phi, node_theta))

    if add_node_latitude:
        # Using the cos of theta.
        # From 1. (north pole) to -1 (south pole).
        node_features.append(np.cos(node_theta))

    if add_node_longitude:
        # Using the cos and sin, which is already normalized.
        node_features.append(np.cos(node_phi))
        node_features.append(np.sin(node_phi))

    if not node_features:
        node_features = np.zeros([num_nodes, 0], dtype=dtype)
    else:
        node_features = np.stack(node_features, axis=-1)

    # Computing some edge features.
    edge_features = []

    if add_relative_positions:
        relative_position = get_relative_position_in_receiver_local_coordinates(
            node_phi=node_phi,
            node_theta=node_theta,
            senders=senders,
            receivers=receivers,
            latitude_local_coordinates=relative_latitude_local_coordinates,
            longitude_local_coordinates=relative_longitude_local_coordinates,
        )

        # Note this is L2 distance in 3d space, rather than geodesic distance.
        relative_edge_distances = np.linalg.norm(relative_position, axis=-1, keepdims=True)

        # Normalize to the maximum edge distance. Note that we expect to always
        # have an edge that goes in the opposite direction of any given edge
        # so the distribution of relative positions should be symmetric around
        # zero. So by scaling by the maximum length, we expect all relative
        # positions to fall in the [-1., 1.] interval, and all relative distances
        # to fall in the [0., 1.] interval.
        max_edge_distance = relative_edge_distances.max()
        edge_features.append(relative_edge_distances / max_edge_distance)
        edge_features.append(relative_position / max_edge_distance)

    if not edge_features:
        edge_features = np.zeros([num_edges, 0], dtype=dtype)
    else:
        edge_features = np.concatenate(edge_features, axis=-1)

    if sine_cosine_encoding:

        def sine_cosine_transform(x: np.ndarray) -> np.ndarray:
            freqs = encoding_multiplicative_factor ** np.arange(encoding_num_freqs)
            phases = freqs * x[..., None]
            x_sin = np.sin(phases)
            x_cos = np.cos(phases)
            x_cat = np.concatenate([x_sin, x_cos], axis=-1)
            return x_cat.reshape([x.shape[0], -1])

        node_features = sine_cosine_transform(node_features)
        edge_features = sine_cosine_transform(edge_features)

    return node_features, edge_features


def lat_lon_to_leading_axes(grid_xarray: xarray.DataArray) -> xarray.DataArray:
    """Reorders xarray so lat/lon axes come first."""
    # leading + ["lat", "lon"] + trailing
    # to
    # ["lat", "lon"] + leading + trailing
    return grid_xarray.transpose("lat", "lon", ...)


def restore_leading_axes(grid_xarray: xarray.DataArray) -> xarray.DataArray:
    """Reorders xarray so batch/time/level axes come first (if present)."""

    # ["lat", "lon"] + [(batch,) (time,) (level,)] + trailing
    # to
    # [(batch,) (time,) (level,)] + ["lat", "lon"] + trailing

    input_dims = list(grid_xarray.dims)
    output_dims = list(input_dims)
    for leading_key in ["level", "time", "batch"]:  # reverse order for insert
        if leading_key in input_dims:
            output_dims.remove(leading_key)
            output_dims.insert(0, leading_key)
    return grid_xarray.transpose(*output_dims)


def lat_lon_deg_to_spherical(
    node_lat: np.ndarray,
    node_lon: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """Convert lat and lon to spherical coordiantes."""
    phi = np.deg2rad(node_lon)
    theta = np.deg2rad(90 - node_lat)
    return phi, theta


def spherical_to_lat_lon(
    phi: np.ndarray,
    theta: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """Convert spherical coordinates to lat and lon."""
    lon = np.mod(np.rad2deg(phi), 360)
    lat = 90 - np.rad2deg(theta)
    return lat, lon


def cartesian_to_spherical(
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """Convert cartesian coordinates to spherical."""
    phi = np.arctan2(y, x)
    with np.errstate(invalid="ignore"):  # circumventing b/253179568
        theta = np.arccos(z)  # Assuming unit radius.
    return phi, theta


def spherical_to_cartesian(
    phi: np.ndarray, theta: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Convert spherical coordinates to cartesian."""
    # Assuming unit radius.
    return (np.cos(phi) * np.sin(theta), np.sin(phi) * np.sin(theta), np.cos(theta))


def get_relative_position_in_receiver_local_coordinates(
    node_phi: np.ndarray,
    node_theta: np.ndarray,
    senders: np.ndarray,
    receivers: np.ndarray,
    latitude_local_coordinates: bool,
    longitude_local_coordinates: bool,
) -> np.ndarray:
    """Returns relative position features for the edges.

    The relative positions will be computed in a rotated space for a local
    coordinate system as defined by the receiver. The relative positions are
    simply obtained by subtracting sender position minues receiver position in
    that local coordinate system after the rotation in R^3.

    Args:
      node_phi: [num_nodes] with polar angles.
      node_theta: [num_nodes] with azimuthal angles.
      senders: [num_edges] with indices.
      receivers: [num_edges] with indices.
      latitude_local_coordinates: Whether to rotate edges such that in the
          positions are computed such that the receiver is always at latitude 0.
      longitude_local_coordinates: Whether to rotate edges such that in the
          positions are computed such that the receiver is always at longitude 0.

    Returns:
      Array of relative positions in R3 [num_edges, 3]
    """

    node_pos = np.stack(spherical_to_cartesian(node_phi, node_theta), axis=-1)

    # No rotation in this case.
    if not (latitude_local_coordinates or longitude_local_coordinates):
        return node_pos[senders] - node_pos[receivers]

    # Get rotation matrices for the local space space for every node.
    rotation_matrices = get_rotation_matrices_to_local_coordinates(
        reference_phi=node_phi,
        reference_theta=node_theta,
        rotate_latitude=latitude_local_coordinates,
        rotate_longitude=longitude_local_coordinates,
    )

    # Each edge will be rotated according to the rotation matrix of its receiver
    # node.
    edge_rotation_matrices = rotation_matrices[receivers]

    # Rotate all nodes to the rotated space of the corresponding edge.
    # Note for receivers we can also do the matmul first and the gather second:
    # ```
    # receiver_pos_in_rotated_space = rotate_with_matrices(
    #    rotation_matrices, node_pos)[receivers]
    # ```
    # which is more efficient, however, we do gather first to keep it more
    # symmetric with the sender computation.
    receiver_pos_in_rotated_space = rotate_with_matrices(
        edge_rotation_matrices, node_pos[receivers]
    )
    sender_pos_in_in_rotated_space = rotate_with_matrices(edge_rotation_matrices, node_pos[senders])
    # Note, here, that because the rotated space is chosen according to the
    # receiver, if:
    # * latitude_local_coordinates = True: latitude for the receivers will be
    #   0, that is the z coordinate will always be 0.
    # * longitude_local_coordinates = True: longitude for the receivers will be
    #   0, that is the y coordinate will be 0.

    # Now we can just subtract.
    # Note we are rotating to a local coordinate system, where the y-z axes are
    # parallel to a tangent plane to the sphere, but still remain in a 3d space.
    # Note that if both `latitude_local_coordinates` and
    # `longitude_local_coordinates` are True, and edges are short,
    # then the difference in x coordinate between sender and receiver
    # should be small, so we could consider dropping the new x coordinate if
    # we wanted to the tangent plane, however in doing so
    # we would lose information about the curvature of the mesh, which may be
    # important for very coarse meshes.
    return sender_pos_in_in_rotated_space - receiver_pos_in_rotated_space


def get_rotation_matrices_to_local_coordinates(
    reference_phi: np.ndarray,
    reference_theta: np.ndarray,
    rotate_latitude: bool,
    rotate_longitude: bool,
) -> np.ndarray:
    """Returns a rotation matrix to rotate to a point based on a reference vector.

    The rotation matrix is build such that, a vector in the
    same coordinate system at the reference point that points towards the pole
    before the rotation, continues to point towards the pole after the rotation.

    Args:
      reference_phi: [leading_axis] Polar angles of the reference.
      reference_theta: [leading_axis] Azimuthal angles of the reference.
      rotate_latitude: Whether to produce a rotation matrix that would rotate
          R^3 vectors to zero latitude.
      rotate_longitude: Whether to produce a rotation matrix that would rotate
          R^3 vectors to zero longitude.

    Returns:
      Matrices of shape [leading_axis] such that when applied to the reference
          position with `rotate_with_matrices(rotation_matrices, reference_pos)`

          * phi goes to 0. if "rotate_longitude" is True.

          * theta goes to np.pi / 2 if "rotate_latitude" is True.

          The rotation consists of:
          * rotate_latitude = False, rotate_longitude = True:
              Latitude preserving rotation.
          * rotate_latitude = True, rotate_longitude = True:
              Latitude preserving rotation, followed by longitude preserving
              rotation.
          * rotate_latitude = True, rotate_longitude = False:
              Latitude preserving rotation, followed by longitude preserving
              rotation, and the inverse of the latitude preserving rotation. Note
              this is computationally different from rotating the longitude only
              and is. We do it like this, so the polar geodesic curve, continues
              to be aligned with one of the axis after the rotation.

    """

    if rotate_longitude and rotate_latitude:
        # We first rotate around the z axis "minus the azimuthal angle", to get the
        # point with zero longitude
        azimuthal_rotation = -reference_phi

        # One then we will do a polar rotation (which can be done along the y
        # axis now that we are at longitude 0.), "minus the polar angle plus 2pi"
        # to get the point with zero latitude.
        polar_rotation = -reference_theta + np.pi / 2

        return transform.Rotation.from_euler(
            "zy", np.stack([azimuthal_rotation, polar_rotation], axis=1)
        ).as_matrix()
    elif rotate_longitude:
        # Just like the previous case, but applying only the azimuthal rotation.
        azimuthal_rotation = -reference_phi
        return transform.Rotation.from_euler("z", -reference_phi).as_matrix()
    elif rotate_latitude:
        # Just like the first case, but after doing the polar rotation, undoing
        # the azimuthal rotation.
        azimuthal_rotation = -reference_phi
        polar_rotation = -reference_theta + np.pi / 2

        return transform.Rotation.from_euler(
            "zyz", np.stack([azimuthal_rotation, polar_rotation, -azimuthal_rotation], axis=1)
        ).as_matrix()
    else:
        raise ValueError("At least one of longitude and latitude should be rotated.")


def rotate_with_matrices(rotation_matrices: np.ndarray, positions: np.ndarray) -> np.ndarray:
    """Rotate with matrices."""
    return np.einsum("bji,bi->bj", rotation_matrices, positions)


def get_bipartite_graph_spatial_features(
    *,
    senders_node_lat: np.ndarray,
    senders_node_lon: np.ndarray,
    senders: np.ndarray,
    receivers_node_lat: np.ndarray,
    receivers_node_lon: np.ndarray,
    receivers: np.ndarray,
    add_node_positions: bool,
    add_node_latitude: bool,
    add_node_longitude: bool,
    add_relative_positions: bool,
    edge_normalization_factor: Optional[float] = None,
    relative_longitude_local_coordinates: bool,
    relative_latitude_local_coordinates: bool,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Computes spatial features for the nodes.

    This function is almost identical to `get_graph_spatial_features`. The only
    difference is that sender nodes and receiver nodes can be in different arrays.
    This is necessary to enable combination with typed Graph.

    Args:
      senders_node_lat: Latitudes in the [-90, 90] interval of shape
        [num_sender_nodes]
      senders_node_lon: Longitudes in the [0, 360] interval of shape
        [num_sender_nodes]
      senders: Sender indices of shape [num_edges], indices in [0,
        num_sender_nodes)
      receivers_node_lat: Latitudes in the [-90, 90] interval of shape
        [num_receiver_nodes]
      receivers_node_lon: Longitudes in the [0, 360] interval of shape
        [num_receiver_nodes]
      receivers: Receiver indices of shape [num_edges], indices in [0,
        num_receiver_nodes)
      add_node_positions: Add unit norm absolute positions.
      add_node_latitude: Add a feature for latitude (cos(90 - lat)) Note even if
        this is set to False, the model may be able to infer the longitude from
        relative features, unless `relative_latitude_local_coordinates` is also
        True, or if there is any bias on the relative edge sizes for different
        longitudes.
      add_node_longitude: Add features for longitude (cos(lon), sin(lon)). Note
        even if this is set to False, the model may be able to infer the longitude
        from relative features, unless `relative_longitude_local_coordinates` is
        also True, or if there is any bias on the relative edge sizes for
        different longitudes.
      add_relative_positions: Whether to relative positions in R3 to the edges.
      edge_normalization_factor: Allows explicitly controlling edge normalization.
        If None, defaults to max edge length. This supports using pre-trained
        model weights with a different graph structure to what it was trained on.
      relative_longitude_local_coordinates: If True, relative positions are
        computed in a local space where the receiver is at 0 longitude.
      relative_latitude_local_coordinates: If True, relative positions are
        computed in a local space where the receiver is at 0 latitude.

    Returns:
      Arrays of shape: [num_nodes, num_features] and [num_edges, num_features].
      with node and edge features.

    """

    num_senders = senders_node_lat.shape[0]
    num_receivers = receivers_node_lat.shape[0]
    num_edges = senders.shape[0]
    dtype = senders_node_lat.dtype
    assert receivers_node_lat.dtype == dtype
    senders_node_phi, senders_node_theta = lat_lon_deg_to_spherical(
        senders_node_lat, senders_node_lon
    )
    receivers_node_phi, receivers_node_theta = lat_lon_deg_to_spherical(
        receivers_node_lat, receivers_node_lon
    )

    # Computing some node features.
    senders_node_features = []
    receivers_node_features = []
    if add_node_positions:
        # Already in [-1, 1.] range.
        senders_node_features.extend(spherical_to_cartesian(senders_node_phi, senders_node_theta))
        receivers_node_features.extend(
            spherical_to_cartesian(receivers_node_phi, receivers_node_theta)
        )

    if add_node_latitude:
        # Using the cos of theta.
        # From 1. (north pole) to -1 (south pole).
        senders_node_features.append(np.cos(senders_node_theta))
        receivers_node_features.append(np.cos(receivers_node_theta))

    if add_node_longitude:
        # Using the cos and sin, which is already normalized.
        senders_node_features.append(np.cos(senders_node_phi))
        senders_node_features.append(np.sin(senders_node_phi))

        receivers_node_features.append(np.cos(receivers_node_phi))
        receivers_node_features.append(np.sin(receivers_node_phi))

    if not senders_node_features:
        senders_node_features = np.zeros([num_senders, 0], dtype=dtype)
        receivers_node_features = np.zeros([num_receivers, 0], dtype=dtype)
    else:
        senders_node_features = np.stack(senders_node_features, axis=-1)
        receivers_node_features = np.stack(receivers_node_features, axis=-1)

    # Computing some edge features.
    edge_features = []

    if add_relative_positions:
        relative_position = get_bipartite_relative_position_in_receiver_local_coordinates(  # pylint: disable=line-too-long
            senders_node_phi=senders_node_phi,
            senders_node_theta=senders_node_theta,
            receivers_node_phi=receivers_node_phi,
            receivers_node_theta=receivers_node_theta,
            senders=senders,
            receivers=receivers,
            latitude_local_coordinates=relative_latitude_local_coordinates,
            longitude_local_coordinates=relative_longitude_local_coordinates,
        )

        # Note this is L2 distance in 3d space, rather than geodesic distance.
        relative_edge_distances = np.linalg.norm(relative_position, axis=-1, keepdims=True)

        if edge_normalization_factor is None:
            # Normalize to the maximum edge distance. Note that we expect to always
            # have an edge that goes in the opposite direction of any given edge
            # so the distribution of relative positions should be symmetric around
            # zero. So by scaling by the maximum length, we expect all relative
            # positions to fall in the [-1., 1.] interval, and all relative distances
            # to fall in the [0., 1.] interval.
            edge_normalization_factor = relative_edge_distances.max()

        edge_features.append(relative_edge_distances / edge_normalization_factor)
        edge_features.append(relative_position / edge_normalization_factor)

    if not edge_features:
        edge_features = np.zeros([num_edges, 0], dtype=dtype)
    else:
        edge_features = np.concatenate(edge_features, axis=-1)

    return senders_node_features, receivers_node_features, edge_features


def get_bipartite_relative_position_in_receiver_local_coordinates(
    senders_node_phi: np.ndarray,
    senders_node_theta: np.ndarray,
    senders: np.ndarray,
    receivers_node_phi: np.ndarray,
    receivers_node_theta: np.ndarray,
    receivers: np.ndarray,
    latitude_local_coordinates: bool,
    longitude_local_coordinates: bool,
) -> np.ndarray:
    """Returns relative position features for the edges.

    This function is equivalent to
    `get_relative_position_in_receiver_local_coordinates`, but adapted to work
    with bipartite typed graphs.

    The relative positions will be computed in a rotated space for a local
    coordinate system as defined by the receiver. The relative positions are
    simply obtained by subtracting sender position minues receiver position in
    that local coordinate system after the rotation in R^3.

    Args:
      senders_node_phi: [num_sender_nodes] with polar angles.
      senders_node_theta: [num_sender_nodes] with azimuthal angles.
      senders: [num_edges] with indices into sender nodes.
      receivers_node_phi: [num_sender_nodes] with polar angles.
      receivers_node_theta: [num_sender_nodes] with azimuthal angles.
      receivers: [num_edges] with indices into receiver nodes.
      latitude_local_coordinates: Whether to rotate edges such that in the
        positions are computed such that the receiver is always at latitude 0.
      longitude_local_coordinates: Whether to rotate edges such that in the
        positions are computed such that the receiver is always at longitude 0.

    Returns:
      Array of relative positions in R3 [num_edges, 3]
    """

    senders_node_pos = np.stack(
        spherical_to_cartesian(senders_node_phi, senders_node_theta), axis=-1
    )

    receivers_node_pos = np.stack(
        spherical_to_cartesian(receivers_node_phi, receivers_node_theta), axis=-1
    )

    # No rotation in this case.
    if not (latitude_local_coordinates or longitude_local_coordinates):
        return senders_node_pos[senders] - receivers_node_pos[receivers]

    # Get rotation matrices for the local space space for every receiver node.
    receiver_rotation_matrices = get_rotation_matrices_to_local_coordinates(
        reference_phi=receivers_node_phi,
        reference_theta=receivers_node_theta,
        rotate_latitude=latitude_local_coordinates,
        rotate_longitude=longitude_local_coordinates,
    )

    # Each edge will be rotated according to the rotation matrix of its receiver
    # node.
    edge_rotation_matrices = receiver_rotation_matrices[receivers]

    # Rotate all nodes to the rotated space of the corresponding edge.
    # Note for receivers we can also do the matmul first and the gather second:
    # ```
    # receiver_pos_in_rotated_space = rotate_with_matrices(
    #    rotation_matrices, node_pos)[receivers]
    # ```
    # which is more efficient, however, we do gather first to keep it more
    # symmetric with the sender computation.
    receiver_pos_in_rotated_space = rotate_with_matrices(
        edge_rotation_matrices, receivers_node_pos[receivers]
    )
    sender_pos_in_in_rotated_space = rotate_with_matrices(
        edge_rotation_matrices, senders_node_pos[senders]
    )
    # Note, here, that because the rotated space is chosen according to the
    # receiver, if:
    # * latitude_local_coordinates = True: latitude for the receivers will be
    #   0, that is the z coordinate will always be 0.
    # * longitude_local_coordinates = True: longitude for the receivers will be
    #   0, that is the y coordinate will be 0.

    # Now we can just subtract.
    # Note we are rotating to a local coordinate system, where the y-z axes are
    # parallel to a tangent plane to the sphere, but still remain in a 3d space.
    # Note that if both `latitude_local_coordinates` and
    # `longitude_local_coordinates` are True, and edges are short,
    # then the difference in x coordinate between sender and receiver
    # should be small, so we could consider dropping the new x coordinate if
    # we wanted to the tangent plane, however in doing so
    # we would lose information about the curvature of the mesh, which may be
    # important for very coarse meshes.
    return sender_pos_in_in_rotated_space - receiver_pos_in_rotated_space


def variable_to_stacked(
    variable: xarray.Variable,
    sizes: Mapping[str, int],
    preserved_dims: Tuple[str, ...] = ("batch", "lat", "lon"),
) -> xarray.Variable:
    """Converts an xarray.Variable to preserved_dims + ("channels",).

    Any dimensions other than those included in preserved_dims get stacked into a
    final "channels" dimension. If any of the preserved_dims are missing then they
    are added, with the data broadcast/tiled to match the sizes specified in
    `sizes`.

    Args:
      variable: An xarray.Variable.
      sizes: Mapping including sizes for any dimensions which are not present in
        `variable` but are needed for the output. This may be needed for example
        for a static variable with only ("lat", "lon") dims, or if you want to
        encode just the latitude coordinates (a variable with dims ("lat",)).
      preserved_dims: dimensions of variable to not be folded in channels.

    Returns:
      An xarray.Variable with dimensions preserved_dims + ("channels",).
    """
    stack_to_channels_dims = [d for d in variable.dims if d not in preserved_dims]
    if stack_to_channels_dims:
        variable = variable.stack(channels=stack_to_channels_dims)
    dims = {dim: variable.sizes.get(dim) or sizes[dim] for dim in preserved_dims}
    dims["channels"] = variable.sizes.get("channels", 1)
    return variable.set_dims(dims)


def dataset_to_stacked(
    dataset: xarray.Dataset,
    sizes: Optional[Mapping[str, int]] = None,
    preserved_dims: Tuple[str, ...] = ("batch", "lat", "lon"),
) -> xarray.DataArray:
    """Converts an xarray.Dataset to a single stacked array.

    This takes each consistuent data_var, converts it into BHWC layout
    using `variable_to_stacked`, then concats them all along the channels axis.

    Args:
      dataset: An xarray.Dataset.
      sizes: Mapping including sizes for any dimensions which are not present in
        the `dataset` but are needed for the output. See variable_to_stacked.
      preserved_dims: dimensions from the dataset that should not be folded in
        the predictions channels.

    Returns:
      An xarray.DataArray with dimensions preserved_dims + ("channels",).
      Existing coordinates for preserved_dims axes will be preserved, however
      there will be no coordinates for "channels".
    """
    data_vars = [
        variable_to_stacked(dataset.variables[name], sizes or dataset.sizes, preserved_dims)
        for name in sorted(dataset.data_vars.keys())
    ]
    coords = {dim: coord for dim, coord in dataset.coords.items() if dim in preserved_dims}
    return xarray.DataArray(data=xarray.Variable.concat(data_vars, dim="channels"), coords=coords)


def stacked_to_dataset(
    stacked_array: xarray.Variable,
    template_dataset: xarray.Dataset,
    preserved_dims: Tuple[str, ...] = ("batch", "lat", "lon"),
) -> xarray.Dataset:
    """The inverse of dataset_to_stacked.

    Requires a template dataset to demonstrate the variables/shapes/coordinates
    required.
    All variables must have preserved_dims dimensions.

    Args:
      stacked_array: Data in BHWC layout, encoded the same as dataset_to_stacked
        would if it was asked to encode `template_dataset`.
      template_dataset: A template Dataset (or other mapping of DataArrays)
        demonstrating the shape of output required (variables, shapes,
        coordinates etc).
      preserved_dims: dimensions from the target_template that were not folded in
        the predictions channels. The preserved_dims need to be a subset of the
        dims of all the variables of template_dataset.

    Returns:
      An xarray.Dataset (or other mapping of DataArrays) with the same shape and
      type as template_dataset.
    """
    unstack_from_channels_sizes = {}
    var_names = sorted(template_dataset.keys())
    for name in var_names:
        template_var = template_dataset[name]
        if not all(dim in template_var.dims for dim in preserved_dims):
            raise ValueError(
                f"stacked_to_dataset requires all Variables to have {preserved_dims} "
                f"dimensions, but found only {template_var.dims}."
            )
        unstack_from_channels_sizes[name] = {
            dim: size for dim, size in template_var.sizes.items() if dim not in preserved_dims
        }

    channels = {
        name: np.prod(list(unstack_sizes.values()), dtype=np.int64)
        for name, unstack_sizes in unstack_from_channels_sizes.items()
    }
    total_expected_channels = sum(channels.values())
    found_channels = stacked_array.sizes["channels"]
    if total_expected_channels != found_channels:
        raise ValueError(
            f"Expected {total_expected_channels} channels but found "
            f"{found_channels}, when trying to convert a stacked array of shape "
            f"{stacked_array.sizes} to a dataset of shape {template_dataset}."
        )

    data_vars = {}
    index = 0
    for name in var_names:
        template_var = template_dataset[name]
        var = stacked_array.isel({"channels": slice(index, index + channels[name])})
        index += channels[name]
        var = var.unstack({"channels": unstack_from_channels_sizes[name]})
        var = var.transpose(*template_var.dims)
        data_vars[name] = xarray.DataArray(
            data=var,
            coords=template_var.coords,
            # This might not always be the same as the name it's keyed under; it
            # will refer to the original variable name, whereas the key might be
            # some alias e.g. temperature_850 under which it should be logged:
            name=template_var.name,
        )
    return type(template_dataset)(data_vars)  # pytype:disable=not-callable,wrong-arg-count



"""Noise generation utils."""

import einops
import numpy as np
import torch
import torch_harmonics as th


def generate_isotropic_noise(num_lon: int, num_lat: int, num_samples=1, isotropic=True):
    """Generate noise on the grid.

    When isotropic is True it samples the equivalent of white noise on a sphere and project it onto
    a grid using Driscoll and Healy, 1994, algorithm. The power spectrum is normalized to have
    variance 1. We need to assume lons = 2 * lats or lons = 2 * (lats -1). If isotropic is false, it
    samples flat normal random noise.

    Args:
        num_lon (int): number of longitudes in the grid.
        num_lat (int): number of latitudes in the grid.
        num_samples (int): number of indipendent samples. Defaults to 1.
        isotropic (bool): if true generates isotropic noise, else flat noise. Defaults to True.

    Returns:
        grid: Numpy array with shape shape(grid) x num_samples.
    """
    if isotropic:
        if 2 * num_lat == num_lon:
            extend = False
        elif 2 * (num_lat - 1) == num_lon:
            extend = True
        else:
            raise ValueError(
                "Isotropic noise requires grid's shape to be 2N x N or 2N x (N+1): "
                f"got {num_lon} x {num_lat}. If the shape is correct, please specify"
                "isotropic=False in the constructor.",
            )

    if isotropic:
        lmax = num_lat - 1 if extend else num_lat
        mmax = lmax + 1
        coeffs = torch.randn(num_samples, lmax, mmax, dtype=torch.complex64) / np.sqrt(
            (num_lat**2) // 2
        )
        isht = th.InverseRealSHT(
            nlat=num_lat, nlon=num_lon, lmax=lmax, mmax=mmax, grid="equiangular"
        )
        noise = isht(coeffs) * np.sqrt(2 * np.pi)
        noise = einops.rearrange(noise, "b lat lon -> lon lat b").numpy()
    else:
        noise = np.random.randn(num_lon, num_lat, num_samples)
    return noise


def sample_noise_level(sigma_min=0.02, sigma_max=88, rho=7):
    """Generate random sample of noise level.

    Sample a noise level according to the distribution described in the paper.
    Notice that the default values are valid only for training and need to be
    modified for sampling.

    Args:
        sigma_min (float, optional): Defaults to 0.02.
        sigma_max (int, optional): Defaults to 88.
        rho (int, optional): Defaults to 7.

    Returns:
        noise_level: single sample of noise level.
    """
    u = np.random.random()
    noise_level = (
        sigma_max ** (1 / rho) + u * (sigma_min ** (1 / rho) - sigma_max ** (1 / rho))
    ) ** rho
    return noise_level


class Preconditioner(torch.nn.Module):
    """Collection of preconditioning functions.

    These functions are described in Karras (2022), table 1.
    """

    def __init__(self, sigma_data: float = 1):
        """Initialize the preconditioning functions.

        Args:
            sigma_data (float): Karras suggests 0.5, GenCast 1. Defaults to 1.
        """
        super().__init__()
        self.sigma_data = sigma_data

    def c_skip(self, sigma):
        """Scaling factor for skip connection."""
        return self.sigma_data**2 / (sigma**2 + self.sigma_data**2)

    def c_out(self, sigma):
        """Scaling factor for output."""
        return sigma * self.sigma_data / torch.sqrt(sigma**2 + self.sigma_data**2)

    def c_in(self, sigma):
        """Scaling factor for input."""
        return 1 / torch.sqrt(sigma**2 + self.sigma_data**2)

    def c_noise(self, sigma):
        """Scaling factor for noise level."""
        return 1 / 4 * torch.log(sigma)
